/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/CTypeObject.java, ras, dumpviewer 1.6
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 2002
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */
/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin  Description
 * ------  ----   ------  ----------------------------------------------------
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION: Objects corresponding to JVM C type definitions
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.lang.*;



public class CTypeObject {

    DvAddress targetAddr;
    DvAddressSpace addrSpace;
    long value;
    boolean valueRead;
    byte[] data;
    CType type;

    /**
     * Return CType of object
     *
     * @param   none
     * @return  CType 
     * @exception  none
     */
    public CType getType() {
        return type;
    }

    /**
     * Return address of object
     *
     * @param   none
     * @return  address of object as a long integer
     * @exception  none
     */
    public long getAddr() {
        return targetAddr.getAddressAsLong();
    }

    /**
     * Return value of an object of primitive or pointer type
     * Endianness conversion is done if required.
     *
     * @param   none
     * @return  Value of the object as long
     * @exception  DvAddressException for illegal object addresses,
     *             NumberFormatException if size > sizeof(long)
     */
    public long getValue() throws DvAddressException {
        if (type.size > 8)  {
            throw new NumberFormatException();
        }
        if (valueRead)
            return value;
        value = readValue(targetAddr, type.size);
        valueRead = true;
        return value;
    }

    private long readValue(DvAddress addr, int size) throws DvAddressException {
        long val;
        if (size == 1)
            val = (long)addrSpace.readByte(addr);
        else if (size == 2)
            val = (long)addrSpace.readShort(addr);
        else if (size == 4)
            val = (long)addrSpace.readInt(addr);
        else if (size == 8)
            val = (long)addrSpace.readLong(addr);
        else
            val = 0;
        return val;
    }

    /**
     * Constructor  for CTypeObject
     *
     * @param   thisType   CType of object
     * @param   address    address of object
     * @param   as         address space from which data for the object is
     *                     obtained
     * @exception  none
     */
    public CTypeObject(CType thisType, DvAddress address, DvAddressSpace as) {
        type = thisType.getType();
        targetAddr = address;
        addrSpace = as;
        valueRead = false;
    }

    /**
     * Constructor  for CTypeObject
     *
     * @param   typeName   CType name of object
     * @param   address    address of object
     * @param   as         address space from which data for the object is
     *                     obtained
     * @exception  none
     */
    public CTypeObject(String typeName, DvAddress address, DvAddressSpace as) {
        type = CType.find(typeName).getType();
        targetAddr = address;
        addrSpace = as;
        valueRead = false;
    }

    /**
     * Constructor  for CTypeObject
     *
     * @param   thisType   CType of object
     * @param   address    address of object as long
     * @param   as         address space from which data for the object is
     *                     obtained
     * @exception  none
     */

    public CTypeObject(CType thisType, long address, DvAddressSpace as) {
        type = thisType.getType();
        targetAddr = new DvAddress(address);
        addrSpace = as;
        valueRead = false;
    }

    /**
     * Constructor  for CTypeObject
     *
     * @param   typeName   CType name of object
     * @param   address    address of object as long
     * @param   as         address space from which data for the object is
     *                     obtained
     * @exception  none
     */
    public CTypeObject(String typeName, long address, DvAddressSpace as) {
        type = CType.find(typeName).getType();
        targetAddr = new DvAddress(address);
        addrSpace = as;
        valueRead = false;
    }

    /**
     * Constructor  for primitive CTypeObject
     *
     * @param   thisType   CType of object
     * @param   val        value of primitive
     * @exception  none
     */
    public CTypeObject(CType thisType, long val ) {
        type = thisType.getType();
        if (type.baseType != CType.TYPE_PRIMITIVE)
            throw new NumberFormatException();
        valueRead = true;
        value = val;
    }

    /**
     * Return a new object with the specified type. This is equivalent
     * to typecasting in C, but no checks are done to see if the types
     * are compatible.  Note that the original type is lost, when an
     * object is typecast.
     *
     * @param   c          CType to typecast to
     * @return  Typecasted object 
     * @exception  none
     */
    public CTypeObject cast(CType c) {
        type = c.getType();
        return this;
    }


    /**
     * Return a new object corresponding to an object field (by index) 
     *
     * @param   index    field index
     * @return  New CTypeObject corresponding to a field
     * @exception  none
     */
    public CTypeObject getField(int index) {
        try {
            if (index <  type.numFields) {
                CType fieldType = type.getFieldType(index).getType();
                DvAddress fieldAddr = new DvAddress(targetAddr.getAddressAsLong());
                fieldAddr.offsetBy(type.getFieldOffset(index));
                CTypeObject field = new CTypeObject(fieldType, fieldAddr, addrSpace);
                return field;
            }
        } catch (DvAddressException e) {
            //ignore exceptions for now
        }
        return null;

    }

    private DvAddress getFieldAddr(String fieldName) {
        try {
            DvAddress fieldAddr = new DvAddress(targetAddr.getAddressAsLong());
            fieldAddr.offsetBy(type.getFieldOffset(fieldName));
            return fieldAddr;
        } catch (DvAddressException e) {
            return null;
        }
    }

    /**
     * Return a new object corresponding to an object field (by name) 
     * Multiple field names including dereferences can be specified.
     * Syntax :
     *   fieldName          equivalent C 
     *    field1            this.field
     *    ->field1          this->field1
     *    field1.field2     this.field1.field2
     *    field1->field2    this.field1->field2
     *
     *
     * @param   fieldName field name(s)
     * @return  New CTypeObject corresponding to a field
     * @exception  none
     */
    public CTypeObject getField(String fieldName) {
        CType t = type.getType();
        if (t.baseType != CType.TYPE_STRUCT && t.baseType != CType.TYPE_UNION)
            return null;
        int dotIndex;
        int arrowIndex;
        dotIndex = fieldName.indexOf('.');
        arrowIndex = fieldName.indexOf('-');

        if (arrowIndex == 0)
            return deReference().getField(fieldName.substring(2, fieldName.length()));
        if (dotIndex == 0)
            return getField(fieldName.substring(1, fieldName.length()));

        if (arrowIndex > 0 && (dotIndex <= 0 || arrowIndex < dotIndex)) {
            String field1Name = fieldName.substring(0, arrowIndex);
            String field2Name = fieldName.substring(arrowIndex+2);
            try {
                DvAddress fieldAddr = addrSpace.readPointer(getFieldAddr(field1Name));
                CTypeObject field1 =  new CTypeObject(t.getFieldType(field1Name), fieldAddr, addrSpace).deReference();
                return field1.getField(field2Name);
            } catch (DvAddressException e) {
                  return null;
            }

        } else if (dotIndex > 0) {
            String field1Name = fieldName.substring(0, dotIndex);
            String field2Name = fieldName.substring(dotIndex+1);
            DvAddress fieldAddr = getFieldAddr(field1Name);
            CTypeObject field1 =  new CTypeObject(t.getFieldType(field1Name), fieldAddr, addrSpace);
            return field1.getField(field2Name);

        } else {
            DvAddress fieldAddr = getFieldAddr(fieldName);
            return new CTypeObject(t.getFieldType(fieldName), fieldAddr, addrSpace);
        }
    }

    /**
     * Return a new CTypeObject corresponding to an array element
     *
     * @param   index   index of array element
     * @return  CTypeObject corresponding to array element
     * @exception  DvAddressException for illegal array addresses
     */
    public CTypeObject getElement(int index)  throws DvAddressException {
        DvAddress elementAddr = new DvAddress(targetAddr.getAddressAsLong());
        elementAddr.offsetBy(type.arrayInfo.type.size*index);
        return new CTypeObject(type.arrayInfo.type, elementAddr, addrSpace);
    }

    /**
     * Dereference an object pointer ( equivalent to *ptr in C)
     *
     * @param   none
     * @return  CTypeObject corresponding to dereferenced object
     * @exception  none
     */
    public CTypeObject deReference() {
        if (type.derefType != null ) {
            try {
                DvAddress derefAddr = addrSpace.readPointer(targetAddr);
                if (derefAddr.getAddressAsLong() == 0)
                    return null;
                return new CTypeObject(type.derefType, derefAddr, addrSpace);
            } catch (DvAddressException e) {
                    return null;
            }
        }
        return null;
    }

    /**
     * Dereference an object pointer and cast it (by type)
     * ( equivalent to *(newType *)ptr in C)
     *
     * @param   c   Type of dereferenced object
     * @return  CTypeObject corresponding to dereferenced object
     * @exception  none
     */
    public CTypeObject deReference(CType c) {
        if (type.derefType != null ) {
            try {
                DvAddress derefAddr = addrSpace.readPointer(targetAddr);
                if (derefAddr.getAddressAsLong() == 0)
                    return null;
                return new CTypeObject(c, derefAddr, addrSpace);
            } catch (DvAddressException e) {
                    return null;
            }
        }
        return null;
    }

    /**
     * Dereference an object pointer and cast it (by name)
     * ( equivalent to *(newType *)ptr in C)
     *
     * @param   typeName   Type name of dereferenced object
     * @return  CTypeObject corresponding to dereferenced object
     * @exception  none
     */
    public CTypeObject deReference(String typeName) {
        if (type.derefType != null ) {
            try {
                DvAddress derefAddr = addrSpace.readPointer(targetAddr);
                if (derefAddr.getAddressAsLong() == 0)
                    return null;
                CType c = CType.find(typeName);
                if (c == null)
                    return null;
                return new CTypeObject(c, derefAddr, addrSpace);
            } catch (DvAddressException e) {
                    return null;
            }
        }
        return null;
    }

    /**
     * Convert object to readable string
     *
     * @param   none
     * @return  Multi-line string with object names and values
     * @exception  none
     */
    public String toString() {
        try {
            return toString(0);
        } catch (Exception e) {
            // ignore exceptions for now.
            return "Cannot print object";
        }
    }

    /* Add tabs to string; <tabLen> specifies number of tabs to add */
    private String addTabs(String str, int tabLen) {
        for (int i = 0; i < tabLen; i++)
            str += "  ";
        return str;
    }

    /* Read NULL terminated C string from <address> */
    private String cstring(DvAddress address) throws DvAddressException{
        StringBuffer str = new StringBuffer();
        DvAddress ptr = addrSpace.readPointer(address);
        while (true) {
            byte[] c = addrSpace.readBytes(ptr, 1);
            if (c == null || c[0] == '\0')
                break;
            str.append((char)(c[0]));
            ptr.offsetBy(1);
        }
       return new String(str);
    }

    private String addressString(DvAddress addr) {
        if (addrSpace.pointerSize() == 4)
            return Integer.toHexString((int)addr.getAddressAsLong());
        else
            return Long.toHexString(addr.getAddressAsLong());
    }

    /* Convert CTypeObject to string. <tabLen> specifies number of tabs 
       at each newline
     */
    private String toString(int tabLen) throws DvAddressException {
        if (data == null)
            data = addrSpace.readBytes(targetAddr, type.getSize());

        CType t = type;
        String str = "";
        t = t.getType();
        str = addTabs(str, tabLen);
        if (tabLen == 0)
            str += t.getName();
        str = str + " @ 0x"+addressString(targetAddr) +"\n";
        tabLen++;

        if (t.baseType < 0) {
            str = addTabs(str, tabLen);
            getValue();
            str = "value = "+value;
            return str;

        } else if (t.baseType == CType.TYPE_ARRAY ||
                   t.baseType == CType.TYPE_POINTER  ||
                   t.baseType == CType.TYPE_FUNCTION ) {
            // Address has already been printed
        } else if (t.baseType == CType.TYPE_STRUCT || t.baseType == CType.TYPE_UNION) {
            for ( int i = 0; i < t.numFields; i++ ) {
                long val = 0;
                CType f = t.getFieldType(i).getType();
                DvAddress fieldAddr = getFieldAddr(t.getFieldName(i));

                if (f.baseType == CType.TYPE_STRUCT || f.baseType == CType.TYPE_UNION) {
                    CTypeObject fieldObj = new CTypeObject(t.getFieldType(i), fieldAddr, addrSpace);
                    str = addTabs(str, tabLen);
                    str += t.getFieldName(i)+" : ";
                    str += fieldObj.toString(tabLen);

                } else if (f.baseType == CType.TYPE_ARRAY) {
                    str = addTabs(str, tabLen);
                    str += t.getFieldName(i)+ " @ ";
                    str += "0x"+addressString(targetAddr);
                    str += "\n";

                } else if (t.getFieldSize(i) <= 8) {
                    CType ftype = t.getFieldType(i).getType();
                    str = addTabs(str, tabLen);
                    str += t.getFieldName(i)+ " = ";
                    try {
                        if (ftype.baseType == CType.TYPE_POINTER && f.derefType.name != null && ftype.derefType.name.equals("char")) {
                            str += cstring(fieldAddr);
                        } else {
                            int start = t.getFieldOffset(i);
                            val = readValue(fieldAddr, t.getFieldSize(i));
                            str += "0x"+addressString(new DvAddress(val));
                        }
                    } catch (Exception e) {
                        str += "Invalid address " + addressString(fieldAddr);
                    }
                    str += "\n";
                } else {
                    System.out.println("not printing "+t.getFieldName(i)+" type "+f.baseType +" index "+t.getFieldType(i).typeNum);
                }
            }
        }
        return str;
    }

}

