/*
 * @(#)src/classes/sov/com/ibm/jvm/dump/DvFileHistory.java, ras, dumpviewer 1.1
 * ===========================================================================
 * IBM Confidential
 * OCO Source Materials
 *
 * IBM Java(tm)2 SDK, Standard Edition, v 1.3.1
 * (C) Copyright IBM Corp. 1998, 2001.
 * The source code for this program is not published or otherwise divested of
 * its trade secrets, irrespective of what has been deposited with the U.S.
 * Copyright office.
 * ===========================================================================
 */

/*
 *
 * ===========================================================================
 *
 *
 *
 * ===========================================================================
 * Change activity:
 *
 * Reason  Date   Origin   Description
 * ------  ----   ------   ----------------------------------------------------
 * 36524   190702 mpartrid New for feature 36524
 *
 * ===========================================================================
 * Module Information:
 *
 * DESCRIPTION:  Provides a file history mechanism for the File menu of a
 *               parent frame.
 * ===========================================================================
 */

package com.ibm.jvm.dump;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

public class DvFileHistory {

   private static final int MAX_ITEM_LEN = 50;      
   private static final String FILE_SEPARATOR_STR = System.getProperty("file.separator");
   private static String historySerFile; 
   private static int max_itemnames; 
   private static ArrayList itemnameHistory = new ArrayList(max_itemnames); 
   private static ArrayList pathnameHistory = new ArrayList(max_itemnames);

   private IFileHistory caller;
   private JMenu fileMenu;

   // --- IFileHistory interface ----------------------------------------------

   /*
    * Interface that must be implemented by a GUI applcation frame
    * that wants to use the FileHistory class.
    * 
    */
    
        public static interface IFileHistory {

        /**
        *   <p>Description:     Get the application name to identify the configuration
        *                       file in the USER_HOME directory. This name should be
        *                       unique in this directory.
        *   </p>
        *   <p>Input      :     None
        *   </p>
        *   <p>Returned   :     the application name
        *   </p>
        */

        public String getApplicationName();

        /**
        *   <p>Description:     Get a handle to the frame's file menu.
        *   </p>
        *   <p>Input      :     None
        *   </p>
        *   <p>Returned   :    the frame's file menu
        *   </p>
        */

        public JMenu getFileMenu();

        /**
        *   <p>Description:     Perform the load file activity.
        *   </p>
        *   <p>Input      :     None
        *   </p>
        *   <p>Returned   :     the pathname of the loaded file
        *   </p>
        */

        public void loadFile(String pathname);
    }

    // -------------------------------------------------------------------------

    /**
     *   <p>Description:    CONSTRUCTOR: caller is the parent frame that hosts
     *                      the file menu
     *   </p>
     *   <p>Input      :    File history interface.
     *   </p>
     *   <p>Returned   :    None
     *   </p>
     */
   
    public DvFileHistory(IFileHistory caller) {

        this.caller = caller;
        historySerFile = System.getProperty("user.home")
                            + FILE_SEPARATOR_STR
                            + caller.getApplicationName()
                            + "_FILE_HISTORY.cfg";
        String max_itemnames_str = System.getProperty("itemnames.max", "9");
        try {
            max_itemnames = Integer.parseInt(max_itemnames_str);
        }
        catch (NumberFormatException e) {
            System.err.println(e);
            e.printStackTrace();
            System.exit(1);
        }
        if (max_itemnames < 1) {
            max_itemnames = 9;
        }
        fileMenu = caller.getFileMenu();      
    }

    /**
     *   <p>Description:    Initialize itemname and pathname arraylists
     *                      from historySerFile. Build up the additional
     *                      entries in the File menu.
     *   </p>
     *   <p>Input      :    None.
     *   </p>
     *   <p>Returned   :    None.
     *   </p>
     */

    public final void initFileMenuHistory() {

        if (new File(historySerFile).exists()) {
            try {
                FileInputStream fis = new FileInputStream(historySerFile);         
                ObjectInputStream ois = new ObjectInputStream(fis);
                int itemnameCount = ois.readInt();

                // if the user has reduced filehistory.size in the past: cut last items

                if (itemnameCount > max_itemnames) {
                    itemnameCount = max_itemnames;
                }

                if (itemnameCount > 0) {
                    fileMenu.addSeparator();
                }

                for (int i=0; i<itemnameCount; i++) {
                    itemnameHistory.add((String)ois.readObject());
                    pathnameHistory.add((String)ois.readObject());
                    MenuItemWithFixedTooltip item = new MenuItemWithFixedTooltip((i+1)+": "
                                                  +(String)itemnameHistory.get(i));
                    item.setToolTipText((String)pathnameHistory.get(i));
                    item.addActionListener(new ItemListener(i));
                    fileMenu.add(item);
                }
                ois.close();
                fis.close();         
            }
            catch (Exception e) {
                System.err.println("Trouble reading file history entries: " + e);
                e.printStackTrace();
            }
        }
    }

    /**
     *   <p>Description:    Save itemname and pathname arraylists to
     *                      historySerFile.
     *   </p>
     *   <p>Input      :
     *   </p>
     *   <p>Returned   :
     *   </p>
     */

    public void saveHistoryEntries() {

        try {
            FileOutputStream fos = new FileOutputStream(historySerFile);
            ObjectOutputStream oos = new ObjectOutputStream(fos);
            int itemnameCount = itemnameHistory.size();

            oos.writeInt(itemnameCount);

            for (int i=0; i<itemnameCount; i++) {
                oos.writeObject((String)itemnameHistory.get(i));
                oos.writeObject((String)pathnameHistory.get(i));
            }

            oos.flush();
            oos.close();
            fos.close();
        }
        catch (Exception e) {
            System.err.println("Trouble saving file history entries: " + e);
            e.printStackTrace();
        }
    }

    /**
     *   <p>Description:    Insert the last loaded pathname into the File menu if
     *                      it is not present yet. Only max pathnames are shown
     *                      (the max number can be set in Jmon.ini, default is 9).
     *                      Every item starts with "<i>: ", where <i> is in the
     *                      range [1..max]. The loaded itemname will become item
     *                      number 1 in the list.
     *   </p>
     *   <p>Input      :    Name and path of file to insert.
     *   </p>
     *   <p>Returned   :    None.
     *   </p>
     */

    public final void insertPathname(String pathname) {
      
        for (int k=0; k<pathnameHistory.size();k++) {
            if (((String)pathnameHistory.get(k)).equals(pathname)) {
                int index = fileMenu.getItemCount() - itemnameHistory.size() + k;
                fileMenu.remove(index);
                pathnameHistory.remove(k);
                itemnameHistory.remove(k);

                if (itemnameHistory.isEmpty()) {
                    //JSeparator is the last menu item at (index-1)
                    fileMenu.remove(index-1);
                }
                insertPathname(pathname);
                return;
            }
        }       

        if (itemnameHistory.isEmpty()) {
            fileMenu.addSeparator();
        } else {
            // remove all itemname entries to prepare for re-arrangement
            for (int i=fileMenu.getItemCount()-1, j=0; j < itemnameHistory.size(); i--, j++) {
                fileMenu.remove(i);
            }
        }

        if (itemnameHistory.size() == max_itemnames) {
            // fileList is full: remove last entry to get space for the first item
            itemnameHistory.remove(max_itemnames-1);
            pathnameHistory.remove(max_itemnames-1);
        }

        itemnameHistory.add(0, getItemname(pathname));
        pathnameHistory.add(0, pathname);

        for (int i=0; i<itemnameHistory.size(); i++) {
            MenuItemWithFixedTooltip item = new MenuItemWithFixedTooltip((i+1)+": "
                                            +(String)itemnameHistory.get(i));
            item.setToolTipText((String)pathnameHistory.get(i));
            item.addActionListener(new ItemListener(i)); 
            fileMenu.add(item);
        }
    }   


    /**
     *   <p>Description:    Return the itemname (abbreviated itemname if necessary)
     *                      to be shown in the file menu open item list.
     *                      A maximum of MAX_ITEM_LEN characters is used for the
     *                      itemname because we do not want to make the JMenuItem
     *                      entry too wide.                                 
     *   </p>
     *   <p>Input      :    String to hold the pathname.
     *   </p>
     *   <p>Returned   :    The pathname.
     *   </p>
     */

    protected String getItemname(String pathname) {

        final char FILE_SEPARATOR = FILE_SEPARATOR_STR.charAt(0);
        final int pathnameLen = pathname.length();

        // if the pathame is short enough: return whole pathname

        if (pathnameLen <= MAX_ITEM_LEN) {
             return pathname;           
        }

        // if we have only one directory: return whole pathname
        // we will not cut to MAX_ITEM_LEN here

        if (pathname.indexOf(FILE_SEPARATOR_STR) == pathname.lastIndexOf(FILE_SEPARATOR_STR)) {
            return pathname;
        }
        else {
            // abbreviate pathanme: Windows OS like solution

            final int ABBREVIATED_PREFIX_LEN = 6; // e.g.: C:\..\
            final int MAX_PATHNAME_LEN = MAX_ITEM_LEN - ABBREVIATED_PREFIX_LEN;         
            int firstFileSeparatorIndex = 0;

            for (int i=pathnameLen-1; i>=(pathnameLen-MAX_PATHNAME_LEN); i--) {

                if (pathname.charAt(i) == FILE_SEPARATOR) {
                    firstFileSeparatorIndex = i;
                }
            }

            if (firstFileSeparatorIndex > 0) {
                return pathname.substring(0, 3) + ".." 
                        + pathname.substring(firstFileSeparatorIndex, pathnameLen);
            }
            else {
                return pathname.substring(0, 3) + ".." + FILE_SEPARATOR_STR + ".."
                        + pathname.substring(pathnameLen-MAX_PATHNAME_LEN, pathnameLen);
            }
        }
    }

    // --- Helper classes ------------------------------------------------------

    /**
     *   <p>Description:    Create a tooltip location directly over the menu item,
     *                      ie, left allign the tooltip text in "overlay" technique.
     */

    private final class MenuItemWithFixedTooltip extends JMenuItem {

        public MenuItemWithFixedTooltip(String text) {
            super(text);
        }

        public Point getToolTipLocation(MouseEvent e) {
            Graphics g = getGraphics();
            FontMetrics metrics = g.getFontMetrics(g.getFont());
            String prefix = itemnameHistory.size() <= 9 ? "8: " : "88: "; 
            int prefixWidth = metrics.stringWidth(prefix);
            int x = JButton.TRAILING + JButton.LEADING -1 + prefixWidth;
            return new Point(x, 0);
        }
    }

    /**
     *   <p>Description:    Listen to menu item selections.
     *   </p>
     */

    private final class ItemListener implements ActionListener {
        int itemNbr;

        ItemListener(int itemNbr) {
            this.itemNbr = itemNbr;
        }

        public void actionPerformed(ActionEvent e) {
            caller.loadFile((String)pathnameHistory.get(itemNbr));
            JMenuItem item = (JMenuItem)e.getSource();
            DvFileHistory.this.insertPathname(item.getToolTipText());
        }
    }
}
 
/* END OF FILE */
