package com.ibm.jvm.findroots;


import java.util.*;
import java.io.*;
import java.text.*;
import com.ibm.jvm.util.*;


/**
 *  There is a one-to-one mapping between vertices and edgeHead, in other words one entry from
 *  vertices plus one entry from edgeHead constitute an object. So why not put them in an object?
 *  Because the overhead of all the extra bytes is too great.
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class BruteGraph {

    IntegerArray vertices = new IntegerArray();
    IntegerArray edgeHead = new IntegerArray();
    IntegerArray edges = new IntegerArray();
    IntegerArray sizes = new IntegerArray();
    IntegerArray widths;
    IntegerArray reachability;
    final static int UNKNOWN = -1;
    final static int IS_ROOT = 0x80000000;
    final static int END_MARKER = 0x40000000;
    boolean resolved;
    boolean active;
    boolean sorted = true;
    int time;
    IntegerArray discoveryTime;
    IntegerArray finishingTime;
    IntegerArray lastTouchTime;
    IntegerArray components;
    //IntegerArray groups;
    //IntegerArray groupSizes;
    IntegerArray dfsOrder;
    int unresolved;
    IntegerArray rootSet;
    PrintClient printClient;
    BruteGraph kernelDAG;
    public static boolean verbose = false;
    static DateFormat df = DateFormat.getTimeInstance();
    boolean isNotBranch[];
    int counts[];
    final boolean optimize = false;
    int saved;
    int times;
    int notimes;
    int maxdepth = 10;
    int maxroots = 7;
    int prune = 100;
    boolean printall = false;
    boolean printed[];
    static boolean xml = false;
    public boolean nofinal = false;
    static String version = "1.1";
    static boolean printedVersion = false;
    boolean sizematters = false;
    boolean simple = false;
    public boolean printstrings = false;
    int totalLength;
    int totalObjects;
    int visitorColour;
    int lastVisitorColour[];
    BruteGraph inverse;

    final void initVisit() {
        visitorColour++;
        if (lastVisitorColour == null || lastVisitorColour.length != size()) {
            lastVisitorColour = new int[size()];
        }
    }

    final boolean visited(int v) {
        return lastVisitorColour[v] == visitorColour;
    }

    final void visit(int v) {
        lastVisitorColour[v] = visitorColour;
    }

    static final String getTime() {
        Calendar c = Calendar.getInstance();

        return df.format(c.getTime());
    }

    public static final void log(String message) {
        if (!printedVersion) {
            printedVersion = true;
            log("findroots version " + version);
        }
        if (xml)
            System.out.println("<log time='" + getTime() + "'>" + message + "</log>");
        else
            System.out.println(getTime() + ": " + message);
    }

    final boolean isRoot(int index) {
        return (edgeHead.get(index) & IS_ROOT) != 0;
    }

    final void setRoot(int index, boolean flag) {
        if (flag)
            edgeHead.put(index, edgeHead.get(index) | IS_ROOT);
        else
            edgeHead.put(index, edgeHead.get(index) & ~IS_ROOT);
    }

    public boolean active() {
        return active;
    }

    public void activate() {
        active = true;
    }

    final int size() {
        return vertices.size();
    }

    final int idToIndex(int id) {
        assert (sorted);
        return vertices.indexOf(id);
    }

    public final void addVertex(int id, int edgeIds[]) {
        addVertex(id, edgeIds, 0);
    }

    public final void addVertex(int id, int edgeIds[], int size) {
        if (!sizematters)
            size = 0;
        resolved = false;
        sorted = false;
        vertices.add(id);
        sizes.add(size);
        totalObjects++;
        totalLength += size;
        if (edgeIds == null || edgeIds.length == 0) {
            edgeHead.add(END_MARKER);
            return;
        }
        edgeHead.add(edges.size());
        for (int i = 0; i < edgeIds.length; i++) {
            if (edgeIds[i] != 0) {
                edges.add(edgeIds[i]);
                if (i == edgeIds.length - 1)
                    edges.add(END_MARKER);
                else
                    edges.add(edges.size() + 1);
            }
        }
    }

    public final void addVertex(int id) {
        addVertex(id, 0);
    }

    public final void addVertex(int id, int size) {
        if (!sizematters)
            size = 0;
        resolved = false;
        sorted = false;
        vertices.add(id);
        sizes.add(size);
        edgeHead.add(END_MARKER);
        totalObjects++;
        totalLength += size;
    }

    public void sort() {
        if (sorted)
            return;
        vertices.sort(new SortListener() {
                public final void swap(int a, int b) {
                    edgeHead.swap(a, b);
                    sizes.swap(a, b);
                    if (reachability != null)
                        reachability.swap(a, b);
                }
            }
        );
        sorted = true;
    }

    public final void addEdge(int fromId, int toId) {
        int fromIndex = idToIndex(fromId);

        if (fromIndex == -1) {
            addVertex(fromId);
            sort();
            fromIndex = idToIndex(fromId);
        }
        if (fromId == toId) {
            return;
        }
        if (idToIndex(toId) == -1) {
            addVertex(toId);
            sort();
            fromIndex = idToIndex(fromId);
        }
        addEdgeByIndex(fromIndex, toId, true);
    }

    public final void addEdgeByIndex(int fromIndex, int toId, boolean check) {
        assert (!resolved);
        int head = getEdgeHead(fromIndex);

        if (check) {
            for (int i = head; i != END_MARKER; i = edges.get(i + 1)) {
                if (edges.get(i) == toId)
                    return;
            }
        }
        int isRoot = edgeHead.get(fromIndex) & IS_ROOT;

        edgeHead.put(fromIndex, isRoot | edges.size());
        edges.add(toId);
        edges.add(head);
    }

    public final int reachability(int id) {
        int v = idToIndex(id);
        int r = reachability.get(v);

        if (r != -1) {
            return r;
        }
        if (printed != null && printed[v]) {
            reachability.put(v, 0);
            return 0;
        }
        if (components == null) {
            reachVisitor.initialize();
            visit(reachVisitor, v);
            r = reachVisitor.result();
        } else {
            int componentId = components.get(v);

            r = kernelDAG.reachability(componentId);
        }
        reachability.put(v, r);
        return r;
    }

    void markPrinted(int index) {
        printed[index] = true;
        if (!simple) {
            int componentId = components.get(index);
            int component = kernelDAG.idToIndex(componentId);

            kernelDAG.printed[component] = true;
        }
    }

    class ReachVisitor extends Visitor {
        int reach;
        void initialize() {
            reach = -1;
            initVisit();
        }

        VisitorResult enterVertex(int vertex) {
            if (printed != null && printed[vertex]) {
                return null;
            }
            if (optimize && !isRoot(vertex) && !isNotBranch[vertex]) {
                reach += counts[vertex];
                saved += counts[vertex] - sizes.get(vertex);
                times++;
            } else {
                reach += sizes.get(vertex);
                notimes++;
            }
            return null;
        }

        int result() {
            return reach;
        }

        boolean continueSearch(int vertex) {
            if (printed != null && printed[vertex]) {
                return false;
            }
            if (optimize && !isRoot(vertex) && !isNotBranch[vertex])
                return false;
            else
                return true;
        }
    }

    ReachVisitor reachVisitor = new ReachVisitor();

    /** Use brute force to calculate the reachability. Used to check the correctness of the algorithm.
     */
    public int bruteReachability(int id) {
        int v = idToIndex(id);
        Visitor visitor = new Visitor() {
                int reach = -1;
                VisitorResult enterVertex(int vertex) {
                    reach++;
                    return null;
                }

                int result() {
                    return reach;
                }
            };

        visit(visitor, v);
        return visitor.result();
    }

    void depthFirstSearch() {
        log("doing depth first search");
        if (dfsOrder == null) {
            dfsOrder = new IntegerArray(size(), 0);
            for (int i = 0; i < size(); i++)
                dfsOrder.put(i, i);
        }
        dfs(new Visitor() {
                final void initialize() {
                    discoveryTime = new IntegerArray(size(), 0);
                    finishingTime = new IntegerArray(size(), 0);
                    if (optimize) {
                        lastTouchTime = new IntegerArray(size(), 0);
                    }
                    time = 0;
                    initVisit();
                }

                final VisitorResult enterVertex(int vertex) {
                    discoveryTime.put(vertex, ++time);
                    return null;
                }

                final VisitorResult leaveVertex(int vertex, VisitorResult result) {
                    finishingTime.put(vertex, ++time);
                    return result;
                }

                final VisitorResult touchVertex(int vertex) {
                    if (optimize)
                        lastTouchTime.put(vertex, ++time);
                    return null;
                }
            }
        );
        log("finished depth first search");
    }

    public void dfs(Visitor visitor) {
        resolve();
        visitor.initialize();
        for (int i = 0; i < dfsOrder.size(); i++) {
            int v = dfsOrder.get(i);

            if (!visited(v)) {
                VisitorResult result = null;
                visitor.root(v, result);
                result = visit(visitor, v);
                visitor.rootAfter(v, result);
            }
        }
    }

    class FindBranchResult extends VisitorResult {
        int before;
        int after;
        int count;
    }

    private final void findBranches() {
        log("enter findBranches");
        depthFirstSearch();
        isNotBranch = new boolean[size()];
        counts = new int[size()];
        dfs(new Visitor() {
                final VisitorResult enterVertex(int vertex) {
                    FindBranchResult result = new FindBranchResult();

                    result.before = discoveryTime.get(vertex);
                    result.after = finishingTime.get(vertex);
                    return result;
                }

                final VisitorResult leaveVertex(int vertex, VisitorResult r) {
                    FindBranchResult result = (FindBranchResult) r;

                    if (lastTouchTime.get(vertex) > result.after)
                        result.after = lastTouchTime.get(vertex);
                    return result;
                }

                final VisitorResult touchVertex(int vertex) {
                    FindBranchResult result = new FindBranchResult();

                    result.before = discoveryTime.get(vertex);
                    if (lastTouchTime.get(vertex) > finishingTime.get(vertex))
                        result.after = lastTouchTime.get(vertex);
                    else
                        result.after = finishingTime.get(vertex);
                    return result;
                }
	
                final VisitorResult compareVertex(int vertex, VisitorResult p, VisitorResult c) {
                    FindBranchResult parent = (FindBranchResult) p;
                    FindBranchResult child = (FindBranchResult) c;

                    if (child.before < parent.before) {
                        parent.before = child.before;
                        isNotBranch[vertex] = true;
                    }
                    if (child.after > parent.after) {
                        parent.after = child.after;
                        isNotBranch[vertex] = true;
                    }
                    return parent;
                }
            }
        );
        dfs(new Visitor() {
                final VisitorResult enterVertex(int vertex) {
                    FindBranchResult result = new FindBranchResult();

                    result.count = sizes.get(vertex);
                    return result;
                }

                final VisitorResult compareVertex(int vertex, VisitorResult p, VisitorResult c) {
                    FindBranchResult parent = (FindBranchResult) p;
                    FindBranchResult child = (FindBranchResult) c;

                    if (child != null) {
                        parent.count += child.count;
                    }
                    return parent;
                }

                final VisitorResult leaveVertex(int vertex, VisitorResult r) {
                    FindBranchResult result = (FindBranchResult) r;

                    if (!isNotBranch[vertex]) {
                        counts[vertex] = result.count;
                    }
                    return result;
                }
            }
        );
        log("leave findBranches");
    }

    final int getEdgeHead(int index) {
        int i = edgeHead.get(index);

        if ((i & END_MARKER) != 0)
            return END_MARKER;
        else
            return i & ~IS_ROOT;
    }

    private final VisitorResult visit(Visitor visitor, int vertex) {
        assert (resolved);
        visit(vertex);
        VisitorResult result = visitor.enterVertex(vertex);

        if (visitor.continueSearch(vertex)) {
            for (int i = getEdgeHead(vertex); i != END_MARKER; i = edges.get(i + 1)) {
                int v = edges.get(i);

                if (v == UNKNOWN) continue;
                VisitorResult childResult;

                if (!visited(v)) {
                    childResult = visit(visitor, v);
                } else {
                    childResult = visitor.touchVertex(v);
                }
                result = visitor.compareVertex(vertex, result, childResult);
            }
        }
        result = visitor.leaveVertex(vertex, result);
        return result;
    }

    private BruteGraph inverse() {
        log("enter inverse");
        resolve();
        BruteGraph graph = new BruteGraph();

        graph.vertices = vertices;
        graph.sizes = null;
        graph.edgeHead = new IntegerArray(size(), END_MARKER);
        for (int from = 0; from < size(); from++) {
            int fromId = vertices.get(from);

            for (int to = getEdgeHead(from); to != END_MARKER; to = edges.get(to + 1)) {
                int toIndex = edges.get(to);

                if (toIndex == UNKNOWN) continue;
                graph.addEdgeByIndex(toIndex, fromId, false);
            }
        }
        graph.dfsOrder = new IntegerArray(size(), 0);
        for (int i = 0; i < size(); i++)
            graph.dfsOrder.put(i, i);
        graph.resolve();
        log("exit inverse");
        return graph;
    }

    public void stronglyConnectedComponents() {
        log("enter stronglyConnectedComponents");
        resolve();
        inverse = inverse();

        inverse.depthFirstSearch();
        inverse.finishingTime.reverseSort(inverse.dfsOrder);
        dfsOrder = inverse.dfsOrder;
        discoveryTime = null;
        finishingTime = null;
        lastTouchTime = null;
        inverse.discoveryTime = null;
        inverse.finishingTime = null;
        inverse.lastTouchTime = null;
        log("finished calculating inverse order");
        /*
        groupSizes = new IntegerArray(1, 0);
        groups = new IntegerArray(size(), -1);
        dfs(new Visitor() {
            final VisitorResult enterVertex(int vertex) {
                int groupIndex = groupSizes.size() - 1;
                int count = groupSizes.get(groupIndex);
                groupSizes.put(groupIndex, count + 1);
                groups.put(vertex, groupIndex);
                for (int i = inverse.getEdgeHead(vertex); i != END_MARKER; i = inverse.edges.get(i + 1)) {
                    int v = inverse.edges.get(i);
                    if (v == UNKNOWN) continue;
                    if (!visited(v)) {
                        visit(this, v);
                    }
                }
                return null;
            }
            final void rootAfter(int vertex, VisitorResult result) {
                groupSizes.add(0);
            }
        });
        log("finished calculating group sizes");
        */
        inverse = null;
        kernelDAG = new BruteGraph();
        components = new IntegerArray(size(), -1);
        Visitor visitor = new Visitor() {
            final VisitorResult enterVertex(int vertex) {
                int index = kernelDAG.size() - 1;

                components.put(vertex, index);
                kernelDAG.sizes.put(index, kernelDAG.sizes.get(index) + sizes.get(vertex) + 1);
                return null;
            }

            final void root(int vertex, VisitorResult result) {
                int index = kernelDAG.size();

                kernelDAG.addVertex(index);
                kernelDAG.setRoot(index, true);
            }
        };

        dfs(visitor);
        log("finished creating kernel DAG vertices");
        kernelDAG.sort();
        log("add edges to kernel DAG");
        for (int i = 0; i < size(); i++) {
            for (int j = getEdgeHead(i); j != END_MARKER; j = edges.get(j + 1)) {
                int v = edges.get(j);

                if (v == UNKNOWN) continue;
                int fromComponent = components.get(i);
                int toComponent = components.get(v);

                if (fromComponent == toComponent)
                    continue;
                assert (fromComponent >= 0 && toComponent >= 0);
                kernelDAG.addEdge(fromComponent, toComponent);
                int toIndex = kernelDAG.idToIndex(toComponent);

                kernelDAG.setRoot(toIndex, false);
            }
        }
        log("finished adding edges to kernel DAG");
        kernelDAG.resolve();
        kernelDAG.dfsOrder = new IntegerArray();
        for (int i = 0; i < kernelDAG.size(); i++) {
            if (kernelDAG.isRoot(i)) {
                kernelDAG.dfsOrder.add(i);
            }
        }
        assert (kernelDAG.unresolved == 0);
        log("exit stronglyConnectedComponents our size = " + size() + " kernel size = " + kernelDAG.size());
        if (optimize)
            kernelDAG.findBranches();
    }

    /** Returns the list of vertices which are in root components. */
    int[] findRoots() {
        if (rootSet == null) {
            log("enter findRoots");
            stronglyConnectedComponents();
            rootSet = new IntegerArray();
            boolean done[] = new boolean[kernelDAG.size()];

            for (int i = 0; i < size(); i++) {
                int c = components.get(i);
                int cindex = kernelDAG.idToIndex(c);

                if (kernelDAG.isRoot(cindex) && !done[c]) {
                    rootSet.add(vertices.get(i));
                    done[c] = true;
                }
            }
            log("exit findRoots, number of roots = " + rootSet.size());
        }
        return rootSet.toArray();
    }

    void resolve() {
        if (resolved) return;
        log("enter resolve");
        resolved = true;
        sort();
        for (int i = 0; i < edges.size(); i += 2) {
            int id = edges.get(i);
            int index = idToIndex(id);

            if (index == -1) {
                unresolved++;
                edges.put(i, UNKNOWN);
            } else {
                edges.put(i, index);
            }
        }
        log("exit resolve");
    }

    class Visitor {

        void initialize() {
            initVisit();
        }

        VisitorResult enterVertex(int vertex) {
            return null;
        }

        VisitorResult leaveVertex(int vertex, VisitorResult result) {
            return result;
        }

        VisitorResult touchVertex(int vertex) {
            return null;
        }

        VisitorResult compareVertex(int vertex, VisitorResult parent, VisitorResult child) {
            return parent;
        }

        void root(int vertex, VisitorResult result) {}
        void rootAfter(int vertex, VisitorResult result) {}

        boolean continueSearch(int vertex) {
            return true;
        }

        int result() {
            return 0;
        }
    }


    class VisitorResult {}

    static BruteGraph random(int seed, int vertices, int edges) {
        Random random = new Random(seed);
        BruteGraph graph = new BruteGraph();

        for (int i = 1; i <= vertices; i++) {
            graph.addVertex(i);
        }
        graph.vertices.sort();
        for (int i = 0; i < edges;) {
            int fromIndex = random.nextInt(vertices);
            int toId = random.nextInt(vertices) + 1;

            if (fromIndex == toId - 1)
                continue;
            graph.addEdgeByIndex(fromIndex, toId, true);
            i++;
        }
        return graph;
    }

    public String toString() {
        StringBuffer s = new StringBuffer();

        for (int i = 0; i < size(); i++) {
            int head = getEdgeHead(i);
            int fromId = vertices.get(i);

            for (int j = head; j != END_MARKER; j = edges.get(j + 1)) {
                int to = edges.get(j);

                if (resolved)
                    to = vertices.get(to);
                s.append(fromId + " --> " + to + "\n");
            }
        }
        return s.toString();
    }

    public static void main(String args[]) {
        log("create random graph");
        int size = Integer.parseInt(args[0]);
        int seed = Integer.parseInt(args[1]);
        int ratio = Integer.parseInt(args[2]);
        BruteGraph graph = random(seed, size, (size*ratio)/(ratio - 1));
        log("finished creating random graph");
        //graph.depthFirstSearch();
        graph.prune = 0;
        graph.print(null);
        System.out.println(graph);
    }

    void check() {
        int roots[] = findRoots();

        for (int i = 0; i < roots.length; i++) {
            //optimize = true;
            //kernelDAG.optimize = true;
            int r1 = reachability(roots[i]);
            //optimize = false;
            //kernelDAG.optimize = false;
            int r2 = reachability(roots[i]);

            if (r1 != r2) {
                System.out.println("at id = " + roots[i] + " r1 = " + r1 + " r2 = " + r2);
                System.out.print(this);
                System.out.println("kernel dag:\n" + kernelDAG);
                //optimize = true;
                //kernelDAG.optimize = true;
                reachability(roots[i]);
                System.out.println("non opt:");
                //optimize = false;
                //kernelDAG.optimize = false;
                reachability(roots[i]);
                System.exit(1);
            }
        }
        System.out.println("ok");
    }

    static String hex(int i) {
        return Integer.toHexString(i);
    }

    static void assert(boolean condition) {
        if (!condition)
            throw new Error("assert failed!");
    }

    public static String[] options() {
        return new String[] {"-depth=<int>", "-roots=<int>|all", "-prune=<int>", "-printall",
                "-xml", "-nofinal", "-version", "-sizematters", "-verbose", "-simple",
                "-printstrings", "-help"};
    }

    public static String[] optionDescriptions() {
        return new String[] {
                "\tDepth of the tree for each root (default 10)",
                "The number of roots to print (default 7)",
                "\tPrune nodes with fewer children (default 100)",
                "\tPrint objects even if they have been done once",
                "\t\tPrint output in XML",
                "\tExclude java/lang/ref/Finalizer",
                "\tPrint version",
                "\tUse object sizes as basis for tree order",
                "\tPrint extra debugging info",
                "\t\tUse simple non-exact algorithm that is much quicker",
                "\tPrint the first 40 chars of arrays of char (SVC dumps only)",
                "\t\tPrint extended help info in html format"
            };
    }

    public boolean parseOption(String arg) {
        if (arg.equals("-version")) {
            System.out.println("Version " + version);
            System.exit(0);
        } else if (arg.equals("-help")) {
            InputStream is = BruteGraph.class.getResourceAsStream("help.html");
            int c;

            try {
                while ((c = is.read()) != -1) {
                    System.out.print((char) c);
                }
            } catch (Exception e) {
                System.err.println("Problem reading help.html");
            }
            System.exit(0);
        } else if ("-printstrings".equals(arg)) {
            printstrings = true;
        } else if ("-simple".equals(arg)) {
            simple = true;
        } else if ("-verbose".equals(arg)) {
            verbose = true;
        } else if ("-sizematters".equals(arg)) {
            sizematters = true;
        } else if ("-nofinal".equals(arg)) {
            nofinal = true;
        } else if ("-xml".equals(arg)) {
            xml = true;
        } else if ("-printall".equals(arg)) {
            printall = true;
        } else {
            int pos = arg.indexOf('=');

            if (pos < 0)
                return false;
            String key = arg.substring(0, pos);
            String value = arg.substring(pos + 1);

            if ("-depth".equals(key)) {
                maxdepth = Integer.parseInt(value, 10);
            } else if ("-roots".equals(key)) {
                if (value.equals("all"))
                    maxroots = -1;
                else
                    maxroots = Integer.parseInt(value, 10);
            } else if ("-prune".equals(key)) {
                prune = Integer.parseInt(value, 10);
            } else
                return false;
        }
        active = true;
        return true;
    }

    void calculateReachability() {
        log("calculate reachability");
        reachability.putAll(-1);
        kernelDAG.reachability.putAll(-1);
        IntegerArray r = new IntegerArray(); 

        int maxReach = -1;
        for (int i = 0; i < rootSet.size(); i++) {
            int id = rootSet.get(i);
            int v = idToIndex(id);
            int reach = 0;
            /*
             * Don't bother with small groups
             */
            //if (groupSizes.get(v) >= maxReach)
                reach = reachability(id);
            if (reach > maxReach)
                maxReach = reach;
            r.add(reach);
        }
        r.reverseSort(rootSet);
        log("finished calculate reachability\n");
    }

    final int width(int vertex) {
        int width = sizematters ? sizes.get(vertex) : 1;

        visit(vertex);
        for (int i = getEdgeHead(vertex); i != END_MARKER; i = edges.get(i + 1)) {
            int v = edges.get(i);

            if (v == UNKNOWN) continue;
            if (!visited(v)) {
                width += width(v);
            }
        }
        widths.put(vertex, width);
        return width;
    }

    void info() {
        if (sizematters)
            log("Now we print the largest roots with the biggest one first up to a maximum of " + maxroots + " roots. The number to the left is the total space occupied by all the objects that are reachable from that object. Child objects are indented.");
        else
            log("Now we print the largest roots with the biggest one first up to a maximum of " + maxroots + " roots. The number to the left is the count of all the objects that are reachable from that object. Child objects are indented.");
    }

    void simplePrint() {
        resolve();
        initVisit();

        widths = new IntegerArray(size(), 0); 
        IntegerArray counts = new IntegerArray(size(), 0); 
        IntegerArray ids = new IntegerArray(); 

        for (int i = 0; i < size(); i++) {
            ids.add(vertices.get(i));
            if (!visited(i)) {
                width(i);
            }
            counts.put(i, widths.get(i));
        }
        counts.reverseSort(ids);
        counts = null;
        if (maxroots == -1)
            maxroots = size();
        info();
        int runningTotal = 0;

        for (int i = 0, done = 0; i < size() && done < maxroots; i++) {
            int id = ids.get(i);
            int vertex = idToIndex(id);

            if (!printed[vertex]) {
                int width = widths.get(vertex);

                if (width < prune) {
                    System.out.println("\nRemaining roots pruned because they are smaller than " + prune);
                    break;
                }
                runningTotal += width;
                if (xml)
                    System.out.println("<root index='" + i + "' running_total = '" + runningTotal + "'>");
                else
                    System.out.println("\n*** root " + i + " (running total " + runningTotal + ") ***\n");
                printVertex(ids.get(i));
                if (xml)
                    System.out.println("</root>");
                done++;
            }
        }
    }

    public void print(PrintClient client) {
        printClient = client;
        printed = new boolean[size()];
        if (simple) {
            simplePrint();
            return;
        }
        findRoots();
        kernelDAG.printed = new boolean[kernelDAG.size()];
        if (maxroots == -1)
            maxroots = rootSet.size();
        info();
        int runningTotal = 0;
        long starttime = System.currentTimeMillis();

        reachability = new IntegerArray(size(), -1);
        kernelDAG.reachability = new IntegerArray(kernelDAG.size(), -1);
        for (int i = 0; i < maxroots; i++) {
            calculateReachability();
            runningTotal += reachability(rootSet.get(0));
            if (xml)
                System.out.println("<root index='" + i + "' running_total = '" + runningTotal + "'>");
            else
                System.out.println("\n*** root " + i + " (running total " + runningTotal + ") ***\n");
            printVertex(rootSet.get(0));
            if (xml)
                System.out.println("</root>");
            long time = System.currentTimeMillis();
        }
        log("total size " + totalLength + " total objects " + totalObjects);
    }

    class PrintVisitorResult extends VisitorResult {
        boolean printed;
    }

    public void printVertex(int id) {
        int index = idToIndex(id);

        dfsOrder = new IntegerArray();
        dfsOrder.add(idToIndex(id));
        dfs(new Visitor() {
                int depth = 0;
                final VisitorResult enterVertex(int vertex) {
                    depth++;
                    PrintVisitorResult result = new PrintVisitorResult();

                    if (printed[vertex] && !printall) {
                        return result;
                    }
                    if (depth >= maxdepth)
                        return result;
                    int count;

                    if (simple) {
                        count = widths.get(vertex);
                    } else {
                        count = reachability(vertices.get(vertex));
                    }
                    if (count < prune)
                        return result;
                    if (!xml) for (int i = 0; i < depth - 1; i++) System.out.print("    ");
                    int id = vertices.get(vertex);
                    String name = (printClient == null) ? "" : printClient.getName(id);

                    if (xml)
                        System.out.println("<node count='" + count + "' id='0x" + hex(id) + "'>" + name);
                    else
                        System.out.println(count + " 0x" + hex(id) + " " + name);
                    result.printed = true;
                    return result;
                }

                final VisitorResult leaveVertex(int vertex, VisitorResult r) {
                    PrintVisitorResult result = (PrintVisitorResult) r;

                    depth--;
                    markPrinted(vertex);
                    if (result.printed && xml)
                        System.out.println("</node>");
                    return result;
                }

                final boolean continueSearch(int vertex) {
                    return printall || !printed[vertex];
                }
            }
        );
    }

    public void printTopOfTree(int id) {
        int index = idToIndex(id);

        dfsOrder = new IntegerArray();
        dfsOrder.add(idToIndex(id));
        dfs(new Visitor() {
            int depth = 0;
            final VisitorResult enterVertex(int vertex) {
                depth++;

                for (int i = 0; i < depth - 1; i++) System.out.print("    ");
                int id = vertices.get(vertex);
                String name = (printClient == null) ? "" : printClient.getName(id);
                System.out.println("0x" + hex(id) + " " + name);

                return null;
            }

            final VisitorResult leaveVertex(int vertex, VisitorResult r) {
                depth--;
                return null;
            }

            final boolean continueSearch(int vertex) {
                return depth < 7;
            }
        });
    }
}
