package com.ibm.jvm.findroots;

import java.io.*;
import java.util.*;
import java.text.*;
import com.ibm.jvm.util.*;
import com.ibm.jvm.svcdump.*;

/**
 *  This class converts a heap dump from native format into Portable Heap Dump (PHD) format.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Convert extends Base implements HeapdumpContentHandler {

    Hashtable classNames = new Hashtable();
    int currentClassId;
    Heapdump dumpWriter = new Heapdump();
    int jvmIndex = -1;
    boolean includeDead = false;

    String className() {
        return "Convert";
    }

    String[] options() {
        return new String[] {"-verbose", "-jvm <jvm index>"};
    }

    String[] optionDescriptions() {
        return new String[] {
            "\tPrint extra debugging info",
            "\tSpecify which jvm to use"
        };
    }

    public boolean parseOption(String arg, String opt) {
        if ("-jvm".equals(arg)) {
            try {
                jvmIndex = Integer.parseInt(opt);
            } catch (Exception e) {
                System.err.println(opt + " is not a valid number");
                usage();
            }
            return true;
        } else if ("-includedead".equals(arg)) {
            includeDead = true;
            return true;
        }
        return false;
    }

    public static void main(String args[]) throws Exception {
        new Convert(args);
    }

    void printChoicesAndExit(Jvm jvms[]) {
        System.err.println("Choose a jvm from the following list:");
        for (int i = 0; i < jvms.length; i++) {
            System.err.println("\tjvm index = " + i + " asid = " + hex(jvms[i].asid()));
        }
        System.exit(1);
    }

    Convert(String args[]) throws Exception {
        parseOptions(args);
        String fileName = args[args.length - 1];

        if (Dump.isSupportedSource(fileName).booleanValue()) {
            Dump dump = new Dump(fileName);
            if (verbose)
                dump.verbose = true;
            Jvm jvm = null;
            Jvm jvms[] = dump.jvms();
            if (jvms.length > 1) {
                if (jvmIndex == -1) {
                    System.err.println("More than one JVM found, please use the -jvm");
                    System.err.println("option to specify which one you want.");
                    printChoicesAndExit(jvms);
                }
                if (jvmIndex >= jvms.length) {
                    System.err.println("Error - maximum value for jvm index = " + (jvms.length-1));
                    System.exit(1);
                }
                jvm = jvms[jvmIndex];
            } else {
                jvm = jvms[0];
            }
            FileOutputStream fos = new FileOutputStream(fileName + ".phd");
            dumpWriter.createOutputStream(fos);
            jvm.scanHeap(this, includeDead);
        } else {
            FileInputStream fis = new FileInputStream(fileName);

            /*
             * figure out if this is hprof (1.3) or javetrace (118) file.
             */
            int c1 = fis.read();
            int c2 = fis.read();
            int c3 = fis.read();

            if (c1 == 'J' && c2 == 'A' && c3 == 'V') {
                // hprof file
                new HprofParseFile(this, fileName);
            } else if ((c1 == '/' && c2 == '/') || (c1 == '0' && c2 == 'x')) {
                FileOutputStream fos = new FileOutputStream(fileName + ".phd");
                dumpWriter.createOutputStream(fos);
                new HeapdumpParser(this, fileName);
            } else {
                fis = new FileInputStream(fileName);
                BufferedInputStream bis = new BufferedInputStream(fis);
                InputStreamReader isr = null;

                if (c1 == '*' && c2 == '*' && c3 == '*') {
                    // ascii javatrace
                    isr = new InputStreamReader(bis, "8859_1");
                } else if (c1 == 0x5c && c2 == 0x5c && c3 == 0x5c) {
                    // ebcdic javatrace
                    isr = new InputStreamReader(bis, "Cp1047");
                } else if (c1 == 'n' && c2 == 'c' && c3 == ' ') {
                    // edited ascii javatrace
                    isr = new InputStreamReader(bis, "8859_1");
                } else {
                    serror("Unknown data format: " + fileName);
                }
                new JavatraceParser(this, isr);
            }
        }
        dumpWriter.close();
    }

    /*
     * HeapdumpContentHandler methods
     */

    int classNameIndex(String className) {
        Integer classId = (Integer)classNames.get(className);
        if (classId == null) {
            classId = new Integer(currentClassId++);
            classNames.put(className, classId);
            dumpWriter.stringDump(className);
        }
        return classId.intValue();
    }

    public void instanceDump(int id, String className, int references[], int size) {
        dumpWriter.objectDump(id, classNameIndex(className), 0, size, references);
    }

    public void classDump(int id, String className, int references[], int size) {
        dumpWriter.classDump(id, classNameIndex(className), 0, size, references);
    }

    public void objectArrayDump(int id, String className, int references[], int size) {
        dumpWriter.objectArrayDump(id, classNameIndex(className), 0, size, references);
    }

    public void primitiveArrayDump(int id, int type, int size) {
        dumpWriter.primitiveArrayDump(id, type, 0, size);
    }

    public void addEdge(int fromId, int toId) {
        error("Unimplemented method");
    }

    static void serror(String msg) {
        throw new Error(msg);
    }

    public void error(String msg) {
        serror(msg);
    }

    static String hex(int i) {
        return Integer.toHexString(i);
    }
}
