package com.ibm.jvm.findroots;


import java.io.*;
import java.util.*;
import java.text.*;
import com.ibm.jvm.util.*;
import com.ibm.jvm.svcdump.*;


/**
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class FindRoots implements HeapdumpContentHandler, PrintClient {

    Graph graph = new Graph();
    Hashtable idToClass = new Hashtable();
    static DateFormat df = DateFormat.getTimeInstance();

    public final void log(String message) {
        Calendar c = Calendar.getInstance();
        String time = df.format(c.getTime());
        System.out.println(time + ": " + message);
    }

    public static void usage() {
        System.err.print("Usage: java com.ibm.jvm.findroots.FindRoots");
        for (int i = 0; i < Graph.options().length; i++)
            System.err.print(" " + Graph.options()[i]);
        System.err.println(" <filename>");
        for (int i = 0; i < Graph.options().length; i++)
            System.err.println("\t" + Graph.options()[i] + "\t" + Graph.optionDescriptions()[i]);
        System.exit(1);
    }

    public static void main(String args[]) throws Exception {
        if (args.length < 1)
            usage();
        FindRoots roots = new FindRoots();

        for (int i = 0; i < args.length; i++) {
            if (args[i].charAt(0) == '-' && !roots.graph.parseOption(args[i]))
                usage();
        }
        if (roots.graph.xml)
            System.out.println("<graph>");
        String fileName = args[args.length - 1];

        if (fileName.endsWith(".digraph")) {
            new DigraphParser(roots, fileName);
            roots.graph.check();
        } else if (Dump.isSupportedSource(fileName).booleanValue()) {
            Dump dump = new Dump(fileName, roots.graph);
            Jvm jvm = dump.jvms()[0];

            roots.graph.activate();
            jvm.findRoots(roots.graph);
            if (roots.graph.xml)
                System.out.println("</graph>");
            return;
        } else {
            FileInputStream fis = new FileInputStream(fileName);

            /*
             * figure out if this is hprof (1.3) or javetrace (118) file.
             */
            int c1 = fis.read();
            int c2 = fis.read();
            int c3 = fis.read();

            if (c1 == 'J' && c2 == 'A' && c3 == 'V') {
                // hprof file
                new HprofParseFile(roots, fileName);
            } else {
                fis = new FileInputStream(fileName);
                BufferedInputStream bis = new BufferedInputStream(fis);
                InputStreamReader isr = null;

                if (c1 == '*' && c2 == '*' && c3 == '*') {
                    // ascii javatrace
                    isr = new InputStreamReader(bis, "8859_1");
                } else if (c1 == 0x5c && c2 == 0x5c && c3 == 0x5c) {
                    // ebcdic javatrace
                    isr = new InputStreamReader(bis, "Cp1047");
                } else if (c1 == 'n' && c2 == 'c' && c3 == ' ') {
                    // edited ascii javatrace
                    isr = new InputStreamReader(bis, "8859_1");
                } else {
                    serror("Unknown data format: " + fileName);
                }
                new JavatraceParser(roots, isr);
            }
        }
        roots.print();
        if (roots.graph.xml)
            System.out.println("</graph>");
    }

    public void print() {
        graph.print(this);
    }

    public String getName(int id) {
        return (String) idToClass.get(new Integer(id));
    }

    public void instanceDump(int id, String className, int references[], int size) {
        idToClass.put(new Integer(id), className);
        if (!graph.nofinal || !className.equals("java/lang/ref/Finalizer")) {
            graph.addVertex(id, references, size);
        }
    }

    public void classDump(int id, String className, int references[], int size) {
        idToClass.put(new Integer(id), className);
        graph.addVertex(id, references, size);
    }

    public void objectArrayDump(int id, String className, int references[], int size) {
        idToClass.put(new Integer(id), "array of " + className);
        graph.addVertex(id, references, 4 * references.length + 8);
    }

    public void primitiveArrayDump(int id, int type, int size) {
        idToClass.put(new Integer(id), "array of type " + type + " and size " + size);
        if (type > 11 || type < 4)
            error("invalid type: " + type);
        graph.addVertex(id, size * (1 << (type & 3)));
    }

    public void addEdge(int fromId, int toId) {
        graph.addEdge(fromId, toId);
    }

    static void serror(String msg) {
        throw new Error(msg);
    }

    public void error(String msg) {
        serror(msg);
    }
}
