package com.ibm.jvm.findroots;

import java.io.*;
import java.util.*;
import com.ibm.jvm.util.*;
import com.ibm.jvm.svcdump.*;

/**
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Heapdump extends Base implements HeapdumpListener {

    DataInputStream dis;
    DataOutputStream dos;
    Vector strings = new Vector();
    int lastId;
    static final int USE_BYTE_SIZES = 0x80;
    static final int USE_SHORT_SIZES = 0x40;
    static final int USE_SHORT_REFS = 0x20;
    static final int IS_LEAF = 0x10;
    static final int TAG_MASK = 0x0f;

    String className() {
        return "Heapdump";
    }

    public void close() {
        try {
            dos.close();
        } catch (Exception e) {
            throw new Error("unexpected error while closing file" + e);
        }
    }

    public void parse(InputStream is, HeapdumpListener listener) {
        try {
            BufferedInputStream bis = new BufferedInputStream(is);
            dis = new DataInputStream(bis);
            try {
                String header = dis.readUTF();
                if (!header.equals("binary heap dump"))
                    throw new Error("invalid header - are you sure this is a .phd file?");
            } catch (Exception e) {
                throw new Error("invalid header - are you sure this is a .phd file?");
            }
            for (;;) {
                int tag = dis.readByte();
                switch (tag & TAG_MASK) {
                case 1:
                    String string = dis.readUTF();
                    strings.addElement(string);
                    listener.stringDump(string);
                    break;
                case 2:
                case 3:
                case 4:
                case 5: {
                    int id, size, refCount = 0;
                    if ((tag & USE_BYTE_SIZES) != 0) {
                        int rid = dis.readByte();
                        size = dis.readByte() & 0xff;
                        if ((tag & IS_LEAF) == 0) {
                            refCount = dis.readByte() & 0xff;
                        }
                        id = lastId + rid;
//System.out.println("tag = " + hex(tag) + " id = " + hex(id) + " lastId = " + hex(lastId) + " rid = " + hex(rid) + " size = " + hex(size));
                    } else if ((tag & USE_SHORT_SIZES) != 0) {
                        int rid = dis.readShort();
                        size = dis.readShort() & 0xffff;
                        if ((tag & IS_LEAF) == 0) {
                            refCount = dis.readShort() & 0xffff;
                        }
                        id = lastId + rid;
//System.out.println("tag = " + hex(tag) + " id = " + hex(id) + " lastId = " + hex(lastId) + " rid = " + hex(rid) + " size = " + hex(size));
                    } else {
                        id = dis.readInt();
                        size = dis.readInt();
                        if ((tag & IS_LEAF) == 0)
                            refCount = dis.readInt();
//System.out.println("tag = " + hex(tag) + " id = " + hex(id) + " lastId = " + hex(lastId) + " size = " + hex(size));
                    }
                    lastId = id;
                    int indexOrType;
                    if ((tag & TAG_MASK) == 5)
                        indexOrType = dis.readByte();
                    else
                        indexOrType = dis.readShort();
                    int refs[] = new int[refCount];
                    for (int i = 0; i < refCount; i++) {
                        if ((tag & USE_SHORT_REFS) == 0) {
                            refs[i] = dis.readInt();
                        } else {
                            int ref = dis.readShort();
                            refs[i] = ref + id;
                        }
                    }
                    switch (tag & TAG_MASK) {
                    case 2:
                        listener.objectDump(id, indexOrType, 0, size, refs);
                        break;
                    case 3:
                        listener.classDump(id, indexOrType, 0, size, refs);
                        break;
                    case 4:
                        listener.objectArrayDump(id, indexOrType, 0, size, refs);
                        break;
                    case 5:
                        listener.primitiveArrayDump(id, indexOrType, 0, size);
                        break;
                    default:
                        assert(false);
                    }
                    break;
                }
                default:
                    throw new Error("unknown type " + tag);
                }
            }
        } catch (EOFException e) {
        } catch (Exception e) {
            e.printStackTrace();
            throw new Error("unexpected error: " + e);
        }
    }

    public void parse(String fileName, HeapdumpListener listener) {
        try {
            parse(new FileInputStream(fileName), listener);
        } catch (Exception e) {
            throw new Error("unexpected error: " + e);
        }
    }

    public void createOutputStream(OutputStream os) {
        try {
            BufferedOutputStream bos = new BufferedOutputStream(os);
            dos = new DataOutputStream(bos);
            dos.writeUTF("binary heap dump");
        } catch (Exception e) {
            throw new Error("unexpected error: " + e);
        }
    }

    public void stringDump(String string) {
        try {
            dos.writeByte(1);
            dos.writeUTF(string);
        } catch (IOException e) {
            throw new Error("unexpected error " + e);
        }
    }

    void instanceDump(int id, int tag, int indexOrType, int size, int refs[]) {
        try {
            /*
             * Set tag flag bits based on the compressions we choose.
             */
            int refCount = (refs == null) ? 0 : refs.length;
            id &= 0xffffffffL;
            if (refCount == 0) {
                tag |= IS_LEAF;
            } else {
                boolean useShortRefs = true;
                for (int i = 0; i < refCount; i++) {
                    int ref = refs[i];
                    if (Math.abs(ref - id) > 0x7fff) {
                        useShortRefs = false;
                    }
                }
                if (useShortRefs)
                    tag |= USE_SHORT_REFS;
            }
            int idGap = id - lastId;
            if (Math.abs(idGap) < 0x80 && size < 0x100 && refCount < 0x100) {
                tag |= USE_BYTE_SIZES;
                dos.writeByte(tag);
                dos.writeByte(idGap);
                dos.writeByte(size);
//System.out.println("tag = " + hex(tag) + " id = " + hex(id) + " lastId = " + hex(lastId) + " rid = " + hex(idGap) + " size = " + hex(size));
                if ((tag & IS_LEAF) == 0)
                    dos.writeByte(refCount);
            } else if (Math.abs(idGap) < 0x8000 && size < 0x10000 && refCount < 0x10000) {
                tag |= USE_SHORT_SIZES;
                dos.writeByte(tag);
                dos.writeShort(idGap);
                dos.writeShort(size);
//System.out.println("tag = " + hex(tag) + " id = " + hex(id) + " lastId = " + hex(lastId) + " rid = " + hex(idGap) + " size = " + hex(size));
                if ((tag & IS_LEAF) == 0)
                    dos.writeShort(refCount);
            } else {
                dos.writeByte(tag);
                dos.writeInt(id);
                dos.writeInt(size);
//System.out.println("tag = " + hex(tag) + " id = " + hex(id) + " lastId = " + hex(lastId) + " size = " + hex(size));
                if ((tag & IS_LEAF) == 0)
                    dos.writeInt(refCount);
            }
            if ((tag & TAG_MASK) == 5)
                dos.writeByte(indexOrType);
            else
                dos.writeShort(indexOrType);
            for (int i = 0; i < refCount; i++) {
                if ((tag & USE_SHORT_REFS) == 0) {
                    dos.writeInt(refs[i]);
                } else {
                    int ref = refs[i];
                    ref -= id;
                    dos.writeShort(ref);
                }
            }
            lastId = id;
        } catch (IOException e) {
            throw new Error("unexpected error " + e);
        }
    }

    public void objectDump(int id, int classStringIndex, int flags, int size, int refs[]) {
        instanceDump(id, 2, classStringIndex, size, refs);
    }

    public void classDump(int id, int nameIndex, int flags, int size, int refs[]) {
        instanceDump(id, 3, nameIndex, size, refs);
    }

    public void objectArrayDump(int id, int classStringIndex, int flags, int size, int refs[]) {
        instanceDump(id, 4, classStringIndex, size, refs);
    }

    public void primitiveArrayDump(int id, int type, int flags, int size) {
        instanceDump(id, 5, type, size, null);
    }
}
