package com.ibm.jvm.findroots;


import java.io.*;
import com.ibm.jvm.util.*;


/**
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class HprofParseFile {

    IntHashtable classes = new IntHashtable();
    HeapdumpContentHandler handler;

    HprofParseFile(HeapdumpContentHandler handler, String filename) throws Exception {
        this.handler = handler;
        parse(filename, new pass1());
        parse(filename, new pass2());
    }

    void parse(String filename, HprofContentHandler handler) throws Exception {
        FileInputStream fis = new FileInputStream(filename);
        BufferedInputStream bis = new BufferedInputStream(fis);

        new HprofParser(handler, bis);
    }

    class pass1 implements HprofContentHandler {

        /*
         * The only thing we do in this pass is to build the class information
         * which we use in the second pass. The main thing we need is the
         * instance field types without which we can't tell where the
         * references are in an object.
         */
        public void loadClass(int objectId, String name) {
            classes.put(objectId, new HprofClass(objectId, name));
        }

        public void classDump(int objectId, int superId, int staticFieldIds[], int instanceFieldTypes[]) {
            HprofClass cl = (HprofClass) classes.get(objectId);

            cl.superClass = (HprofClass) classes.get(superId);
            cl.instanceFieldTypes = instanceFieldTypes;
        }

        public void error(String msg) {
            handler.error(msg);
        }

        // Remaining methods are not used on this pass

        public void header(String header) {}

        public void utf8(int id, String utf8) {}

        public void beginHeapDump() {}

        public void endHeapDump() {}

        public void objectArrayDump(int objectId, int classId, int elements[]) {}

        public void primitiveArrayDump(int objectId, int type, int size) {}

        public void instanceDump(int objectId, int classId, short body[]) {}
    }


    class pass2 implements HprofContentHandler {

        public void classDump(int objectId, int superId, int staticFieldIds[], int instanceFieldTypes[]) {
            HprofClass cl = (HprofClass) classes.get(objectId);

            handler.classDump(objectId, cl.name, staticFieldIds, 300);
        }

        public void instanceDump(int objectId, int classId, short body[]) {
            HprofClass cl = (HprofClass) classes.get(classId);
            String name = cl.name;
            int offset = 0;
            IntegerArray references = new IntegerArray();

            for (; cl != null; cl = cl.superClass) {
                for (int i = 0; i < cl.instanceFieldTypes.length; i++) {
                    if (cl.instanceFieldTypes[i] == 2) {
                        int b1 = body[offset];
                        int b2 = body[offset + 1];
                        int b3 = body[offset + 2];
                        int b4 = body[offset + 3];
                        int id = (b1 << 24) + (b2 << 16) + (b3 << 8) + b4;

                        if (id != 0) {
                            references.add(id);
                        }
                    }
                    offset += (1 << (cl.instanceFieldTypes[i] & 3));
                }
            }
            handler.instanceDump(objectId, name, references.toArray(), body.length + 8);
        }

        public void objectArrayDump(int objectId, int classId, int elements[]) {
            HprofClass cl = (HprofClass) classes.get(classId);

            handler.objectArrayDump(objectId, cl.name, elements, elements.length << 2);
        }

        public void primitiveArrayDump(int objectId, int type, int size) {
            handler.primitiveArrayDump(objectId, type, size);
        }

        public void error(String msg) {
            handler.error(msg);
        }

        // Remaining methods are not used on this pass

        public void loadClass(int objectId, String name) {}

        public void header(String header) {}

        public void utf8(int id, String utf8) {}

        public void beginHeapDump() {}

        public void endHeapDump() {}
    }


    class HprofClass {
        int id;
        String name;
        HprofClass superClass;
        int instanceFieldTypes[];

        HprofClass(int id, String name) {
            this.id = id;
            this.name = name;
        }
    }
}

