package com.ibm.jvm.findroots;


import java.io.*;
import java.util.*;
import com.ibm.jvm.util.*;


/**
 *  Parse a binary Hprof file.
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class HprofParser {

    static final int HPROF_UTF8 = 0x01;
    static final int HPROF_LOAD_CLASS = 0x02;
    static final int HPROF_UNLOAD_CLASS = 0x03;
    static final int HPROF_FRAME = 0x04;
    static final int HPROF_TRACE = 0x05;
    static final int HPROF_ALLOC_SITES = 0x06;
    static final int HPROF_START_THREAD = 0x0a;
    static final int HPROF_END_THREAD = 0x0b;
    static final int HPROF_HEAP_DUMP = 0x0c;
    static final int HPROF_CONTROL_SETTINGS = 0x0e;
    static final int HPROF_GC_ROOT_UNKNOWN = 0xff;
    static final int HPROF_GC_ROOT_JNI_GLOBAL = 0x01;
    static final int HPROF_GC_ROOT_JNI_LOCAL = 0x02;
    static final int HPROF_GC_ROOT_JAVA_FRAME = 0x03;
    static final int HPROF_GC_ROOT_NATIVE_STACK = 0x04;
    static final int HPROF_GC_ROOT_STICKY_CLASS = 0x05;
    static final int HPROF_GC_ROOT_THREAD_BLOCK = 0x06;
    static final int HPROF_GC_ROOT_THREAD_OBJ = 0x08;
    static final int HPROF_GC_CLASS_DUMP = 0x20;
    static final int HPROF_GC_INSTANCE_DUMP = 0x21;
    static final int HPROF_GC_OBJ_ARRAY_DUMP = 0x22;
    static final int HPROF_GC_PRIM_ARRAY_DUMP = 0x23;

    InputStream input;
    HprofContentHandler handler;
    int remainder;				// of bytes in this record
    IntHashtable strings = new IntHashtable();

    HprofParser(HprofContentHandler handler, InputStream input) throws Exception {
        setContentHandler(handler);
        try {
            parse(input);
        } catch (EOFException e) {
            System.out.println("Warning: unexpected EOF, dump may not be complete");
        }
    }

    final int readInputByte() throws Exception {
        int b = input.read();

        if (b == -1)
            throw new EOFException("Unexpected EOF");
        return b;
    }

    void setContentHandler(HprofContentHandler handler) {
        this.handler = handler;
    }

    void parse(InputStream input) throws Exception {
        this.input = input;
        handler.header(readString());
        skip(4);	// size of identifiers - assume 4
        skip(8);	// start time - ignore times for now

        /* Process the records */
        for (;;) {
            int tag = input.read();

            if (tag == -1)
                return;
            skip(4);	// timestamp
            remainder = readInt();

            switch (tag) {
            case HPROF_HEAP_DUMP: {
                    handler.beginHeapDump();
                    while (remainder > 0) {
                        tag = readByte();
                        switch (tag) {
                        case HPROF_GC_CLASS_DUMP: {
                                int objectId = readInt();

                                skip(4);
                                int superId = readInt();

                                skip(5 * 4);
                                int instanceSize = readInt();
                                // skip constant pool
                                int size = readShort();

                                for (int i = 0; i < size; i++) {
                                    skip(2);
                                    int type = readByte();

                                    skip(1 << (type & 3));
                                }
                                size = readShort();	// number of static fields
                                IntegerArray staticFields = new IntegerArray();

                                for (int i = 0; i < size; i++) {
                                    skip(4);		// field name id
                                    int type = readByte();

                                    if (type == 2) {
                                        staticFields.add(readInt());
                                    } else {
                                        skip(1 << (type & 3));
                                    }
                                }
                                size = readShort();	// number of instance fields
                                int instanceFields[] = new int[size];

                                for (int i = 0; i < size; i++) {
                                    skip(4);		// field name id
                                    instanceFields[i] = readByte();
                                }
                                handler.classDump(objectId, superId, staticFields.toArray(), instanceFields);
                                break;
                            }

                        case HPROF_GC_OBJ_ARRAY_DUMP: {
                                int objectId = readInt();

                                skip(4);		// stack trace serial number
                                int size = readInt();
                                int classId = readInt();
                                int elements[] = new int[size];

                                for (int i = 0; i < size; i++) {
                                    elements[i] = readInt();
                                }
                                handler.objectArrayDump(objectId, classId, elements);
if (objectId == 0x49926b88) {
System.out.println("yes we found 0x49926b88!");
System.out.println("classId = " + Integer.toHexString(classId));
//System.exit(1);
}
                                break;
                            }

                        case HPROF_GC_PRIM_ARRAY_DUMP: {
                                int objectId = readInt();

                                skip(4);		// stack trace serial number
                                int size = readInt();
                                int type = readByte();

                                skip(size * (1 << (type & 3)));
                                handler.primitiveArrayDump(objectId, type, size);
                                break;
                            }

                        case HPROF_GC_INSTANCE_DUMP: {
                                int objectId = readInt();

                                skip(4);		// stack trace serial number
                                int classId = readInt();
                                int size = readInt();
                                short body[] = new short[size];

                                for (int i = 0; i < size; i++) {
                                    body[i] = (short) readByte();
                                }
                                handler.instanceDump(objectId, classId, body);
                                break;
                            }

                        case HPROF_GC_ROOT_UNKNOWN:
                            skip(4);
                            break;

                        case HPROF_GC_ROOT_JNI_GLOBAL:
                            skip(8);
                            break;

                        case HPROF_GC_ROOT_JNI_LOCAL:
                            skip(12);
                            break;

                        case HPROF_GC_ROOT_JAVA_FRAME:
                            skip(12);
                            break;

                        case HPROF_GC_ROOT_STICKY_CLASS:
                            skip(4);
                            break;

                        case HPROF_GC_ROOT_THREAD_BLOCK:
                            skip(8);
                            break;

                        case HPROF_GC_ROOT_THREAD_OBJ:
                            skip(12);
                            break;

                        case HPROF_GC_ROOT_NATIVE_STACK:
                            skip(8);
                            break;

                        default:
                            handler.error("Unknown heap dump tag: " + tag);
                            return;
                        }
                    }
                    handler.endHeapDump();
                    break;
                }

            case HPROF_UTF8: {
                    int id = readInt();
                    String string = readString(remainder);

                    strings.put(id, string);
                    handler.utf8(id, string);
                    break;
                }

            case HPROF_LOAD_CLASS: {
                    skip(4);	// class serial number
                    int objectId = readInt();

                    skip(4);	// stack trace serial number
                    int nameId = readInt();
                    String name = (String) strings.get(nameId);

                    handler.loadClass(objectId, name);
System.out.println("load class, id = " + Integer.toHexString(objectId) + ", name = " + name);
                    break;
                }

            case HPROF_UNLOAD_CLASS:
                skip(4);
                break;

            case HPROF_CONTROL_SETTINGS:
                skip(6);
                break;

            case HPROF_START_THREAD:
                skip(6 * 4);
                break;

            case HPROF_END_THREAD:
                skip(4);
                break;

            case HPROF_TRACE: {
                    skip(4);
                    skip(4);
                    int n = readInt();

                    skip(n * 4);
                    break;
                }

            case HPROF_FRAME:
                skip(6 * 4);
                break;

            case HPROF_ALLOC_SITES:
                skip(remainder);
                break;

            default:
                handler.error("Unknown tag: " + tag);
                return;
            }
        }
    }

    final String readString() throws Exception {
        StringBuffer buf = new StringBuffer();
        int c;

        while ((c = readInputByte()) != 0) {
            buf.append((char) c);
        }
        return buf.toString();
    }

    final String readString(int length) throws Exception {
        StringBuffer buf = new StringBuffer();

        for (int i = 0; i < length; i++) {
            buf.append((char) readInputByte());
        }
        return buf.toString();
    }

    final int readInt() throws Exception {
        int b1 = readInputByte();
        int b2 = readInputByte();
        int b3 = readInputByte();
        int b4 = readInputByte();

        remainder -= 4;

        return (b1 << 24) + (b2 << 16) + (b3 << 8) + b4;
    }

    final int readShort() throws Exception {
        int b1 = readInputByte();
        int b2 = readInputByte();

        remainder -= 2;

        return (b1 << 8) + b2;
    }

    final int readByte() throws Exception {
        int b = readInputByte();

        remainder--;

        return b;
    }

    final void skip(int n) throws Exception {
        // Brain-dead InputStream doesn't always skip what you ask it to
        remainder -= n;
        while (n > 0)
            n -= input.skip(n);
    }
}
