package com.ibm.jvm.util;


import java.util.Random;
import java.io.Serializable;


/**
 *  Really useful little utility class that just implements an array of integers that can be sorted. One
 *  main use is for mapping integer addresses to objects without the huge overhead of a Hashtable. The
 *  key to the usefulness is in the SortListener object that gets sorted along with this object.
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public final class IntegerArray implements Serializable, SortListener {
    static final int shift = 12;
    static final int CHUNK_SIZE = 1 << shift;
    static final int mask = CHUNK_SIZE - 1;
    int chunks[][] = new int[1][CHUNK_SIZE];
    int chunkOffset;
    SortListener listener;

    public IntegerArray() {}

    public IntegerArray(int size, int value) {
        for (int i = 0; i < size; i++)
            add(value);
    }

    public Object clone() {
        throw new Error("clone no longer implemented!");
    }

    public int[] toArray() {
        int a[] = new int[size()];

        for (int i = 0; i < chunks.length; i++) {
            System.arraycopy(chunks[i], 0, a, i << shift, i == chunks.length - 1 ? chunkOffset : CHUNK_SIZE);
        }
        return a;
    }

    public void add(int value) {
        if (chunkOffset >= CHUNK_SIZE) {
            int newchunks[][] = new int[chunks.length + 1][];

            for (int i = 0; i < chunks.length; i++) {
                newchunks[i] = chunks[i];
            }
            newchunks[chunks.length] = new int[CHUNK_SIZE];
            chunks = newchunks;
            chunkOffset = 0;
        }
        chunks[chunks.length - 1][chunkOffset++] = value;
    }

    public int size() {
        return ((chunks.length - 1) << shift) + chunkOffset;
    }

    public int get(int index) {
        return chunks[index >> shift][index & mask];
    }

    public void put(int index, int value) {
        chunks[index >> shift][index & mask] = value;
    }

    public void putAll(int value) {
        for (int i = 0; i < size(); i++)
            put(i, value);
    }

    public int indexOf(int value) {
        return indexOf(value, -1);
    }

    public int indexOf(int value, int hint) {
        int low = 0;
        int high = size() - 1;
        if (hint >= 0 && hint <= high) {
            int midVal = get(hint);

            if (midVal < value)
                low = hint + 1;
            else if (midVal > value)
                high = hint - 1;
            else
                return hint;
        }

        while (low <= high) {
            int mid = (low + high) >> 1;
            int midVal = get(mid);

            if (midVal < value)
                low = mid + 1;
            else if (midVal > value)
                high = mid - 1;
            else
                return mid;
        }

        return -1;
    }

    public void sort() {
        sort(null);
    }

    public void reverseSort() {
        reverseSort(null);
    }

    public void sort(SortListener listener) {
        this.listener = listener;
        quickSort(0, size() - 1);
        this.listener = null;
    }

    public void reverseSort(SortListener listener) {
        this.listener = listener;
        quickSort(0, size() - 1);
        for (int i = 0, j = size() - 1; i < j; i++, j--)
            swap(i, j);
        this.listener = null;
    }

    public void swap(int a, int b) {
        int t = get(a);

        put(a, get(b));
        put(b, t);
        if (listener != null)
            listener.swap(a, b);
    }

    void quickSort(int p, int r) {
        if (r - p < 7) {
            for (int i = p; i <= r; i++) {
                for (int j = i; j > p && get(j - 1) > get(j); j--) {
                    swap(j, j - 1);
                }
            }
            return;
        }
        if (p < r) {
            int q = randomPartition(p, r);

            quickSort(p, q);
            quickSort(q + 1, r);
        }
    }

    Random random = new Random();

    int randomPartition(int p, int r) {
        int i = random.nextInt(r - p + 1) + p;

        swap(p, i);
        return partition(p, r);
    }

    int partition(int p, int r) {
        int x = get(p);
        int i = p - 1;
        int j = r + 1;

        while (true) {
            while (get(--j) > x);
            while (get(++i) < x);
            if (i < j)
                swap(i, j);
            else
                return j;
        }
    }
}
