package com.ibm.jvm.util;

import com.ibm.jvm.util.*;
import java.util.BitSet;
import java.io.*;

/**
 * this class provides a very compact representation of sparse bit sets. there are different
 * types of set denoted by the top three bits of root:
 *      EMPTY_SET       does exactly what it says on the tin
 *      ONE_MEMBER_SET  
 */

public final class TreeBitSet implements Serializable {

    static SubsetClass rootSubset;
    static Memory memory = new Memory();
    static BitSet liveRoots = new BitSet();
    static SparseBitSet shared = new SparseBitSet();
    static final int MAX_BITS = 29;
    static final int MAX_BIT = (1 << MAX_BITS) - 1;
    static final int MAX_NBITS = 29;
    static final int MAX_NBIT = (1 << MAX_NBITS) - 1;
    static final int MAX_SMALL_SET_SIZE = 32;
    static final int MAX_MEDIUM_SET_SIZE = 1024;
    static final int EMPTY_SET = 0;
    static final int ONE_MEMBER_SET = 1;
    static final int TWO_MEMBER_SET = 2;
    static final int THREE_MEMBER_SET = 3;
    static final int SMALL_SET = 4;
    static final int MEDIUM_SET = 5;
    static final int LARGE_SET = 6;
    static final int PARTIAL_TREE = -3;
    static final int EMPTY_TREE = -2;
    static final int FULL_TREE = -1;
    static final int COPY_ON_WRITE = 0x80000000;
    static int freeLists[] = new int[256];
    static int freeMemory;
    public static boolean verbose;
    int _root;

    static {
        int i = 7;
        rootSubset = new SubsetClass(null, i);
        if (true) {
        int j = 2;
        for (; i + j < MAX_NBITS; i += j) {
            rootSubset = new SubsetClass(rootSubset, j);
        }
        if (i < MAX_NBITS) {
            rootSubset = new SubsetClass(rootSubset, MAX_NBITS - i);
        }
        } else {
        rootSubset = new SubsetClass(rootSubset, 6);
        rootSubset = new SubsetClass(rootSubset, 5);
        rootSubset = new SubsetClass(rootSubset, 4);
        rootSubset = new SubsetClass(rootSubset, 3);
        rootSubset = new SubsetClass(rootSubset, 3);
        rootSubset = new SubsetClass(rootSubset, 2);
        rootSubset = new SubsetClass(rootSubset, 1);
        }
        int totalBits = rootSubset.shift + rootSubset.bits;
        if (totalBits != MAX_NBITS) 
            throw new Error("total bits is " + totalBits + " but should be " + MAX_NBITS);

        freeLists = new int[2048];
        for (i = 0; i < freeLists.length; i++)
            freeLists[i] = -1;
    }

    public static int freeMemory() {
        return freeMemory * 4;
    }

    public static int usage() {
        return memory.size() * 4;
    }

    public static void reset() {
        memory = new Memory();
        liveRoots = new BitSet();
        for (int i = 0; i < freeLists.length; i++)
            freeLists[i] = -1;
        freeMemory = 0;
    }

    /**
     * construct a new TreeBitSet
     */

    public TreeBitSet() {
        _root = EMPTY_SET;
    }

    static int getType(int root) {
        return root >>> MAX_BITS;
    }

    static int setType(int root, int type) {
        return (root & MAX_BIT) | (type << MAX_BITS);
    }

    /**
     * allocate some memory
     */

    static int alloc(int size) {
        checkFreeList(size);
        int freeList = freeLists[size - 1];
        if (freeList != -1) {
            freeLists[size - 1] = memory.get(freeList);
            for (int i = 0; i < size; i++)
                memory.put(freeList + i, 0);
            freeMemory -= size;
            return freeList;
        }
        int r = memory.size();
        for (int i = 0; i < size; i++)
            memory.add(0);
        return r;
    }

    /**
     * allocate a memory root
     */

    static int rootalloc(int size) {
        int root = alloc(size);
        liveRoots.set(root);
        return root;
    }

    /**
     * make a copy of some memory
     */

    static int copyalloc(int old, int oldSize, int newSize) {
        int r = alloc(newSize);
        for (int i = 0; i < oldSize; i++)
            memory.put(r + i, memory.get(old + i));
        if (liveRoots.get(old))
            liveRoots.set(r);
        return r;
    }

    /**
     * free a memory root
     */

    public static void freeroot(int root, int size) {
        assert(liveRoots.get(root));
        liveRoots.clear(root);
        free(root, size);
    }

    static void checkFreeList(int size) {
        if (size > freeLists.length) {
            int newLists[] = new int[size];
            System.arraycopy(freeLists, 0, newLists, 0, freeLists.length);
            for (int i = freeLists.length; i < newLists.length; i++)
                newLists[i] = -1;
            freeLists = newLists;
        }
    }

    static void free(int r, int size) {
        checkFreeList(size);
        r &= ~COPY_ON_WRITE;
        for (int i = 0; i < size; i++) {
            memory.put(r + i, 0x7fffffff);
        }
        memory.put(r, freeLists[size - 1]);
        freeLists[size - 1] = r;
        freeMemory += size;
    }

    public void set(int bit) {
        _root = set(_root, bit);
    }

    public static int set(int root, int bit) {
        int oldRoot = -1, newRoot = -1;
        assert(bit <= MAX_NBIT);
        if (get(root, bit))
            return root;
        switch (getType(root)) {
        case EMPTY_SET:
            // store the single member directly in the root
            root = setType(bit, ONE_MEMBER_SET);
            break;
        case ONE_MEMBER_SET:
            // allocate some memory to store the two roots
            newRoot = rootalloc(2);
            memory.put(newRoot, root & MAX_BIT);
            memory.put(newRoot + 1, bit);
            root = setType(newRoot, TWO_MEMBER_SET);
            break;
        case TWO_MEMBER_SET:
            // allocate some memory to store the three roots and copy the old two
            oldRoot = root & MAX_BIT;
            assert(liveRoots.get(oldRoot));
            newRoot = copyalloc(oldRoot, 2, 3);
            memory.put(newRoot, memory.get(oldRoot));
            memory.put(newRoot + 1, memory.get(oldRoot + 1));
            memory.put(newRoot + 2, bit);
            freeroot(oldRoot, 2);
            root = setType(newRoot, THREE_MEMBER_SET);
            break;
        case THREE_MEMBER_SET:
            // promote to the more general small set
            oldRoot = root & MAX_BIT;
            assert(liveRoots.get(oldRoot));
            newRoot = rootalloc(5);
            memory.put(newRoot, (4 << 16) | 4);
            memory.put(newRoot + 1, memory.get(oldRoot));
            memory.put(newRoot + 2, memory.get(oldRoot + 1));
            memory.put(newRoot + 3, memory.get(oldRoot + 2));
            memory.put(newRoot + 4, bit);
            freeroot(oldRoot, 3);
            root = setType(newRoot, SMALL_SET);
            break;
        case SMALL_SET:
            oldRoot = root & MAX_BIT;
            assert(liveRoots.get(oldRoot));
            int size = memory.get(oldRoot) & 0xffff;
            int max = memory.get(oldRoot) >>> 16;
            if (size < max) {
                memory.put(oldRoot + size + 1, bit);
                memory.put(oldRoot, (max << 16) | ++size);
            } else {
                max <<= 1;
                if (max <= MAX_SMALL_SET_SIZE) {
                    newRoot = copyalloc(oldRoot, size + 1, max + 1);
                    freeroot(oldRoot, size + 1);
                    memory.put(newRoot + size + 1, bit);
                    memory.put(newRoot, (max << 16) | ++size);
                    root = setType(newRoot, SMALL_SET);
                } else if (MAX_MEDIUM_SET_SIZE > 0) {
                    newRoot = rootalloc((max << 1) + 2);
                    memory.put(newRoot, 0);
                    memory.put(newRoot + 1, max);
                    root = setType(newRoot, MEDIUM_SET);
                    for (int i = 0; i < size; i++) {
                        root = intervalSet(root, memory.get(oldRoot + i + 1));
                    }
                    root = intervalSet(root, bit);
                    freeroot(oldRoot, size + 1);
                } else {
                    int tmp = root;
                    root = promoteToLarge(root);
                    free(tmp);
                }
            }
            break;
        case MEDIUM_SET:
            root = intervalSet(root, bit);
            break;
        case LARGE_SET:
            treeSet(rootSubset, root & MAX_BIT, bit);
            break;
        default:
            assert(false);
        }
        check(root);
        return root;
    }

    /*
    public static int clear(int root, int bit) {
        assert(bit <= MAX_NBIT);
        if (!get(root, bit))
            return root;
        switch (getType(root)) {
        case EMPTY_SET:
            return root;
        case ONE_MEMBER_SET:
            if ((root & MAX_BIT) == bit)
                root = 0;
            return root;
        case TWO_MEMBER_SET:
            if (memory.get(root & MAX_BIT) == bit) {
            }
        case THREE_MEMBER_SET:
        case SMALL_SET:
        case MEDIUM_SET:
            root = intervalClear(root, bit);
            break;
        case LARGE_SET:
            treeClear(rootSubset, root & MAX_BIT, bit);
            break;
        default:
            assert(false);
        }
        check(root);
        return root;
    }
    */


    static int intervalMerge(int root, int lower) {
        int oldRoot = root & MAX_BIT;
        int max = memory.get(oldRoot + 1);
        int intervals = oldRoot + 2;
        int lengths = intervals + max;
        int size = mediumSetSize(root);
        memory.put(lengths + lower, memory.get(lengths +lower) + memory.get(lengths + lower + 1));
        if (lower < size - 2) {
            int len = size - (lower + 2);
            for (int i = 0; i < len; i++) {
                memory.put(intervals + lower + 1 + i, memory.get(intervals + lower + 2 + i));
                memory.put(lengths + lower + 1 + i, memory.get(lengths + lower + 2 + i));
            }
        }
        size--;
        assert(size > 0);
        memory.put(oldRoot, size);
        if (size * 3 < max) {
            int oldMax = max;
            max >>= 1;
            int newRoot = rootalloc((max << 1) + 2);
            int nintervals = newRoot + 2;
            int nlengths = nintervals + max;
            for (int i = 0; i < size; i++) {
                memory.put(nintervals + i, memory.get(intervals + i));
                memory.put(nlengths + i, memory.get(lengths + i));
            }
            memory.put(newRoot, size);
            memory.put(newRoot + 1, max);
            root = setType(newRoot, MEDIUM_SET);
            freeroot(oldRoot, (oldMax << 1) + 2);
        }
        check(root);
        return root;
    }

    static int intervalSet(int root, int bit) {
        assert(getType(root) == MEDIUM_SET);

        int low = 0;
        int size = mediumSetSize(root);
        int high = size - 1;
        int mid = -1;
        int midVal = -1;
        int oldRoot = root & MAX_BIT;
        int intervals = oldRoot + 2;
        int max = memory.get(oldRoot + 1);
        int lengths = intervals + max;

        if (size == 0) {
            assert(max > 0);
            memory.put(oldRoot, 1);
            memory.put(oldRoot + 2, bit);
            memory.put(oldRoot + 2 + max, 1);
            check(root);
            return root;
        }

        while (low <= high) {
            mid = (low + high) >> 1;
            midVal = memory.get(intervals + mid);

            if (bit >= midVal) {
                if (bit < (midVal + memory.get(lengths + mid)))
                    return root;                     // already set
                else if (bit == (midVal + memory.get(lengths + mid))) {
                    memory.put(lengths + mid, memory.get(lengths + mid) + 1);  // add to interval
                    if (mid < (size - 1) && (bit + 1) == memory.get(intervals + mid + 1)) {
                        root = intervalMerge(root, mid);
                    }
                    check(root);
                    return root;
                }
                low = mid + 1;                  // continue search
            } else {
                if (bit == (midVal - 1)) {
                    // add to interval
                    memory.put(intervals + mid, memory.get(intervals + mid) - 1);
                    memory.put(lengths + mid, memory.get(lengths + mid) + 1);
                    if (mid > 0 && bit == (memory.get(intervals + mid - 1) +
                            memory.get(lengths + mid - 1))) {
                        root = intervalMerge(root, mid - 1);
                    }
                    check(root);
                    return root;
                }
                high = mid - 1;                 // continue search
            }
        }
        if (size == max) {
            max <<= 1;
            if (max <= MAX_MEDIUM_SET_SIZE) {
                int newRoot = rootalloc((max << 1) + 2);
                int nintervals = newRoot + 2;
                int nlengths = nintervals + max;
                for (int i = 0; i < size; i++) {
                    memory.put(nintervals + i, memory.get(intervals + i));
                    memory.put(nlengths + i, memory.get(lengths + i));
                }
                memory.put(newRoot, size);
                memory.put(newRoot + 1, max);
                root = setType(newRoot, MEDIUM_SET);
                freeroot(oldRoot, max + 2);
                oldRoot = newRoot;
                intervals = nintervals;
                lengths = nlengths;
            } else {
                root = promoteToLarge(root);
                treeSet(rootSubset, root & MAX_BIT, bit);
                freeroot(oldRoot, max + 2);
                check(root);
                return root;
            }
        }
        if (bit < memory.get(intervals)) {
            for (int i = size - 1; i >= 0; i--) {
                memory.put(intervals + i + 1, memory.get(intervals + i));
                memory.put(lengths + i + 1, memory.get(lengths + i));
            }
            memory.put(intervals, bit);
            memory.put(lengths, 1);
        } else if (bit > memory.get(intervals + size - 1)) {
            memory.put(intervals + size, bit);
            memory.put(lengths + size, 1);
        } else {
            if (bit < midVal)
                mid--;
            assert(size - mid - 2 >= 0);
            for (int i = size - mid - 2; i >= 0; i--) {
                memory.put(intervals + mid + 2 + i, memory.get(intervals + mid + 1 + i));
                memory.put(lengths + mid + 2 + i, memory.get(lengths + mid + 1 + i));
            }
            memory.put(intervals + mid + 1, bit);
            memory.put(lengths + mid + 1, 1);
        }
        memory.put(oldRoot, ++size);
        assert((root & MAX_BIT) == oldRoot);
        check(root);
        return root;
    }

    static int intervalCopy(int oroot) {
        assert(getType(oroot) == MEDIUM_SET);
        int r = oroot & MAX_BIT;
        int size = memory.get(r);
        int max = memory.get(r + 1);
        int root = rootalloc((max << 1) + 2);
        int ointervals = r + 2;
        int olengths = ointervals + max;
        int nintervals = root + 2;
        int nlengths = nintervals + max;
        for (int i = 0; i < size; i++) {
            memory.put(nintervals + i, memory.get(ointervals + i));
            memory.put(nlengths + i, memory.get(olengths + i));
        }
        memory.put(root, size);
        memory.put(root + 1, max);
        return setType(root, MEDIUM_SET);
    }
        
    static int tmpintervals[];
    static int tmplengths[];

    static int intervalOr(int root, int oroot) {
        assert(getType(root) == MEDIUM_SET && getType(oroot) == MEDIUM_SET);
        check(root);
        check(oroot);
        int r = root & MAX_BIT;
        int or = oroot & MAX_BIT;
        int uslen = memory.get(r);
        int themlen = memory.get(or);
        assert(uslen != 0x7fffffff);
        assert(themlen != 0x7fffffff);
        int intervals = r + 2;
        int ointervals = or + 2;
        int max = memory.get(r + 1);
        int lengths = intervals + max;
        int olengths = ointervals + memory.get(or + 1);
        if (tmpintervals == null || tmpintervals.length < (uslen + themlen)) {
            tmpintervals = new int[uslen + themlen];
            tmplengths = new int[uslen + themlen];
        }
        int count = 0, us = 0, them = 0;
        if (memory.get(intervals) <= memory.get(ointervals)) {
            tmpintervals[0] = memory.get(intervals);
            tmplengths[0] = memory.get(lengths);
            us++;
        } else {
            tmpintervals[0] = memory.get(ointervals);
            tmplengths[0] = memory.get(olengths);
            them++;
        }
        for (; us < uslen || them < themlen; ) {
            int intervalsThem = 0x7fffffff, intervalsUs = 0x7fffffff;
            if (us < uslen) {
                intervalsUs = memory.get(intervals + us);
            }
            if (them < themlen) {
                intervalsThem = memory.get(ointervals + them);
            }
            if (us == uslen || (them != themlen && intervalsThem <= intervalsUs)) {
                if (intervalsThem > (tmpintervals[count] + tmplengths[count])) {
                    count++;
                    tmpintervals[count] = intervalsThem;
                    tmplengths[count] = memory.get(olengths + them);
                } else {
                    int tmpmax = tmpintervals[count] + tmplengths[count];
                    int themmax = intervalsThem + memory.get(olengths + them);
                    if (tmpmax > themmax) 
                        tmplengths[count] = tmpmax - tmpintervals[count];
                    else
                        tmplengths[count] = themmax - tmpintervals[count];
                }
                them++;
            } else {
                if (intervalsUs > (tmpintervals[count] + tmplengths[count])) {
                    count++;
                    tmpintervals[count] = intervalsUs;
                    tmplengths[count] = memory.get(lengths + us);
                } else {
                    int tmpmax = tmpintervals[count] + tmplengths[count];
                    int usmax = intervalsUs + memory.get(lengths + us);
                    if (tmpmax > usmax) 
                        tmplengths[count] = tmpmax - tmpintervals[count];
                    else
                        tmplengths[count] = usmax - tmpintervals[count];
                }
                us++;
            }
        }
        count++;
        if (count > max) {
            free(root);
            max = (count * 3) / 2;
            root = rootalloc((max << 1) + 2);
            intervals = root + 2;
            lengths = intervals + max;
            memory.put(root + 1, max);
        }
        for (int i = 0; i < count; i++) {
            memory.put(intervals + i, tmpintervals[i]);
            memory.put(lengths + i, tmplengths[i]);
        }
        memory.put(root & MAX_BIT, count);
        root = setType(root, MEDIUM_SET);
        check(root);
        return root;
    }

    static int promoteToLarge(int root) {
        int newRoot = rootalloc(rootSubset.subsets);
        for (int i = 0; i < rootSubset.subsets; i++) {
            memory.put(newRoot + i, EMPTY_TREE);
        }
        for (IntEnumeration e = new BitEnum(root); e.hasMoreElements(); ) {
            int i = e.nextInt();
            treeSet(rootSubset, newRoot, i);
        }
        return setType(newRoot, LARGE_SET);
    }

    static void treeSet(SubsetClass sc, int subsetRoot, int n) {
        assert((subsetRoot & COPY_ON_WRITE) == 0);
        if (sc.next == null) {
            int subset = n >> 5;
            assert(subset < sc.subsets);
            int newbits = memory.get(subsetRoot + subset) | (1 << (n & 0x1f));
            memory.put(subsetRoot + subset, newbits);
            return;
        }
        int index = sc.subsetIndex(n);
        int nextSubsetRoot = memory.get(subsetRoot + index);
        if (nextSubsetRoot == FULL_TREE)
            return;
        if (nextSubsetRoot == EMPTY_TREE) {
            nextSubsetRoot = treeAlloc(sc.next);
            memory.put(subsetRoot + index, nextSubsetRoot);
        } else {
            if ((nextSubsetRoot & COPY_ON_WRITE) != 0) {
                nextSubsetRoot = treeShallowCopy(sc.next, nextSubsetRoot);
                memory.put(subsetRoot + index, nextSubsetRoot);
            }
        }
        treeSet(sc.next, nextSubsetRoot, sc.subsetBits(n));
        if (treeFull(sc.next, nextSubsetRoot)) {
            treeFree(sc.next, nextSubsetRoot);
            memory.put(subsetRoot + index, FULL_TREE);
        }
    }

    static void treeFree(SubsetClass sc, int subsetRoot) {
        if ((subsetRoot & COPY_ON_WRITE) == 0) {
            if (sc.next != null) {
                for (int i = 0; i < sc.subsets; i++) {
                    int nextSubsetRoot = memory.get(subsetRoot + i);
                    if (nextSubsetRoot != FULL_TREE && nextSubsetRoot != EMPTY_TREE) {
                        treeFree(sc.next, nextSubsetRoot);
                    }
                }
            }
            free(subsetRoot, sc.subsets);
        }
    }

    static int treeMemoryUsage(SubsetClass sc, int subsetRoot) {
        if ((subsetRoot & COPY_ON_WRITE) != 0) {
            subsetRoot &= ~COPY_ON_WRITE;
            if (shared.get(subsetRoot)) {
                return 0;
            }
            shared.set(subsetRoot);
        }
        int usage = sc.subsets;
        if (sc.next != null) {
            for (int i = 0; i < sc.subsets; i++) {
                int nextSubsetRoot = memory.get(subsetRoot + i);
                if (nextSubsetRoot != FULL_TREE && nextSubsetRoot != EMPTY_TREE) {
                    usage += treeMemoryUsage(sc.next, nextSubsetRoot);
                }
            }
        }
        return usage;
    }

    static boolean treeFull(SubsetClass sc, int subsetRoot) {
        assert((subsetRoot & COPY_ON_WRITE) == 0);
        for (int i = 0; i < sc.subsets; i++) {
            /*
             * note: the value of FULL_TREE is chosen to be -1 so that it works for both
             * leaf and branch nodes
             */
            if (memory.get(subsetRoot + i) != FULL_TREE)
                return false;
        }
        return true;
    }

    static boolean treeGet(SubsetClass sc, int subsetRoot, int n) {
        subsetRoot &= ~COPY_ON_WRITE;
        if (sc.next == null) {
            int subset = n >> 5;
            assert(subset < sc.subsets);
            return (memory.get(subsetRoot + subset) & (1 << (n & 0x1f))) != 0;
        }
        int index = sc.subsetIndex(n);
        int nextSubsetRoot = memory.get(subsetRoot + index);
        if (nextSubsetRoot == FULL_TREE)
            return true;
        else if (nextSubsetRoot == EMPTY_TREE)
            return false;
        else
            return treeGet(sc.next, nextSubsetRoot, sc.subsetBits(n));
    }

    static int treeAlloc(SubsetClass sc) {
        int r = alloc(sc.subsets);
        if (sc.next != null) {
            for (int i = 0; i < sc.subsets; i++) {
                memory.put(r + i, EMPTY_TREE);
            }
        }
        return r;
    }

    public boolean get(int bit) {
        return get(_root, bit);
    }

    public static boolean get(int root, int bit) {
        int r;
        assert(bit <= MAX_NBIT);
        switch (getType(root)) {
        case EMPTY_SET:
            return false;
        case ONE_MEMBER_SET:
            return (root & MAX_BIT) == bit;
        case TWO_MEMBER_SET:
            r = root & MAX_BIT;
            return memory.get(r) == bit || memory.get(r + 1) == bit;
        case THREE_MEMBER_SET:
            r = root & MAX_BIT;
            return memory.get(r) == bit || memory.get(r + 1) == bit || memory.get(r + 2) == bit;
        case SMALL_SET:
            r = root & MAX_BIT;
            int size = smallSetSize(root);
            for (int i = 0; i < size; i++)
                if (memory.get(r + i + 1) == bit)
                    return true;
            return false;
        case MEDIUM_SET:
            return intervalGet(root, bit);
        case LARGE_SET:
            return treeGet(rootSubset, root & MAX_BIT, bit);
        default:
            assert(false);
        }
        return false;
    }

    static boolean intervalGet(int root, int bit) {
        assert(getType(root) == MEDIUM_SET);

        int low = 0;
        int size = mediumSetSize(root);
        int high = size - 1;
        int r = root & MAX_BIT;
        int intervals = r + 2;
        int max = memory.get(r + 1);
        int lengths = intervals + max;

        while (low <= high) {
            int mid = (low + high) >> 1;
            int midVal = memory.get(intervals + mid);

            if (bit >= midVal) {
                if (bit < (midVal + memory.get(lengths + mid)))
                    return true;
                low = mid + 1;                  // continue search
            } else {
                high = mid - 1;                 // continue search
            }
        }
        return false;
    }

    static final int treeNextBit(SubsetClass sc, int subsetRoot, int n) {
        subsetRoot &= ~COPY_ON_WRITE;
        if (sc.next == null) {
            int max = sc.subsets;
            for (int subset = n >> 5; subset < max; subset++) {
                int bits = memory.get(subsetRoot + subset);
                for (int i = n & 31; i <= 31; i++) {
                    if (((1 << i) & bits) != 0) {
                        return (subset << 5) + i;
                    }
                }
                n = 0;
            }
            return -1;
        }
        //for (int i = sc.subsetIndex(n); i < sc.subsets; n = (++i) * sc.size) {
        for (int i = sc.subsetIndex(n); i < sc.subsets; n = (++i) << sc.shift) {
            int nextSubsetRoot = memory.get(subsetRoot + i);
            if (nextSubsetRoot == FULL_TREE) {
                //return (i * sc.size) + sc.subsetBits(n);
                return (i << sc.shift) + sc.subsetBits(n);
            } else if (nextSubsetRoot == EMPTY_TREE)
                continue;
            else {
                int j = treeNextBit(sc.next, nextSubsetRoot, sc.subsetBits(n));
                if (j != -1)
                    //return (i * sc.size) + j;
                    return (i << sc.shift) + j;
            }
        }
        return -1;
    }

    static void treeUnion(SubsetClass sc, int usRoot, int themRoot) {
        assert((usRoot & COPY_ON_WRITE) == 0);
        int us = usRoot & ~COPY_ON_WRITE;
        int them = themRoot & ~COPY_ON_WRITE;
        if (sc.next == null) {
            for (int i = 0; i < sc.subsets; i++) {
                memory.put(us + i, memory.get(us + i) | memory.get(them + i));
            }
            return;
        }
        for (int i = 0; i < sc.subsets; i++) {
            int usNext = memory.get(us + i);
            int themNext = memory.get(them + i);
            if (themNext == FULL_TREE) {
                if ((usNext != EMPTY_TREE) && (usNext != FULL_TREE))
                    treeFree(sc.next, usNext);
                memory.put(us + i, FULL_TREE);
            } else if (themNext == EMPTY_TREE)
                continue;
            else if (usNext == FULL_TREE)
                continue;
            else if (usNext == EMPTY_TREE) {
                usNext = treeDeepCopy(sc.next, themNext);
                memory.put(us + i, usNext);
                memory.put(them + i, usNext);
            } else if (treeIsSubset(sc.next, usNext, themNext)) {
                treeFree(sc.next, usNext);
                usNext = treeDeepCopy(sc.next, themNext);
                memory.put(us + i, usNext);
                memory.put(them + i, usNext);
            } else if (treeIsSubset(sc.next, themNext, usNext)) {
                continue;
            } else if (usNext != themNext) {        // may already share a copy
                if ((usNext & COPY_ON_WRITE) != 0) {
                    usNext = treeShallowCopy(sc.next, usNext);
                    memory.put(us + i, usNext);
                }
                treeUnion(sc.next, usNext, themNext);
                if (treeFull(sc.next, usNext)) {
                    treeFree(sc.next, usNext);
                    memory.put(us + i, FULL_TREE);
                }
            }
        }
    }

    static boolean treeIsSubset(SubsetClass sc, int usRoot, int themRoot) {
        int us = usRoot & ~COPY_ON_WRITE;
        int them = themRoot & ~COPY_ON_WRITE;
        if (sc.next == null) {
            for (int i = 0; i < sc.subsets; i++) {
                int u = memory.get(us + i);
                int t = memory.get(them + i);
                if ((u & t) != u) {
                    return false;
                }
            }
            return true;
        }
        for (int i = 0; i < sc.subsets; i++) {
            int usNext = memory.get(us + i);
            int themNext = memory.get(them + i);
            if (themNext == FULL_TREE) {
                continue;
            } else if (themNext == EMPTY_TREE) {
                if (usNext != EMPTY_TREE) {
                    return false;
                }
            } else if (usNext == FULL_TREE) {
                return false;
            } else if (usNext == EMPTY_TREE) {
                continue;
            } else if (usNext != themNext) {
                if (!treeIsSubset(sc.next, usNext, themNext)) {
                    return false;
                }
            }
        }
        return true;
    }

    static int treeDeepCopy(SubsetClass sc, int subsetRoot) {
        subsetRoot &= ~COPY_ON_WRITE;
        if (sc.next != null) {
            for (int i = 0; i < sc.subsets; i++) {
                int nextSubsetRoot = memory.get(subsetRoot + i);
                if (nextSubsetRoot != EMPTY_TREE && nextSubsetRoot != FULL_TREE) {
                    if ((nextSubsetRoot & COPY_ON_WRITE) == 0) {        // optimization
                        nextSubsetRoot = treeDeepCopy(sc.next, nextSubsetRoot);
                        memory.put(subsetRoot + i, nextSubsetRoot);
                    }
                }
            }
        }
        return subsetRoot | COPY_ON_WRITE;
    }

    static int treeShallowCopy(SubsetClass sc, int subsetRoot) {
        subsetRoot &= ~COPY_ON_WRITE;
        int newSubsetRoot = alloc(sc.subsets);
        for (int i = 0; i < sc.subsets; i++) {
            memory.put(newSubsetRoot + i, memory.get(subsetRoot + i));
        }
        return newSubsetRoot;
    }

    static int treePopulation(SubsetClass sc, int subsetRoot) {
        subsetRoot &= ~COPY_ON_WRITE;
        int pop = 0;
        if (sc.next == null) {
            for (int i = 0; i < sc.subsets; i++) {
                int subset = memory.get(subsetRoot + i);
                while (subset != 0) {
                    pop += bytePopulations[subset & 0xff];
                    subset >>>= 8;
                }
            }
            return pop;
        }
        for (int i = 0; i < sc.subsets; i++) {
            int subset = memory.get(subsetRoot + i);
            if (subset == FULL_TREE)
                pop += sc.size;
            else if (subset == EMPTY_TREE)
                continue;
            else
                pop += treePopulation(sc.next, subset);
        }
        return pop;
    }

    static int bytePopulations[] = new int[]
    {
        0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4,
        1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
        1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
        2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
        1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
        2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
        2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
        3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
        1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5,
        2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
        2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
        3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
        2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6,
        3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
        3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7,
        4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
    };

    static int smallSetSize(int root) {
        int r = root & MAX_BIT;
        assert(liveRoots.get(r));
        return memory.get(r) & 0xffff;
    }

    static int mediumSetSize(int root) {
        int r = root & MAX_BIT;
        assert(liveRoots.get(r));
        return memory.get(r);
    }

    public void or(TreeBitSet oset) {
        _root = or(_root, oset._root);
    }

    static void check(int root) {
    }

    public static int close(int root) {
        return root;
    }

    public static int or(int root, int oroot) {
        check(root);
        check(oroot);
        if (root == oroot)
            return root;
        if (getType(oroot) == LARGE_SET) {
            if (getType(root) != LARGE_SET) {
                int r = root;
                root = promoteToLarge(root);
                free(r);
            }
            treeUnion(rootSubset, root & MAX_BIT, oroot & MAX_BIT);
        } else if (getType(oroot) == MEDIUM_SET && getType(root) == LARGE_SET) {
            // note: cannot mix & match the oo and static variants!
            int tmproot = promoteToLarge(oroot);
            treeUnion(rootSubset, root & MAX_BIT, tmproot & MAX_BIT);
            free(tmproot);
        } else if (getType(root) == MEDIUM_SET && getType(oroot) == MEDIUM_SET) {
            root = intervalOr(root, oroot);
        } else if (getType(root) < MEDIUM_SET && getType(oroot) == MEDIUM_SET) {
            int nroot = intervalCopy(oroot);
            nroot = or(nroot, root);
            free(root);
            root = nroot;
        } else {
            for (IntEnumeration e = new BitEnum(oroot); e.hasMoreElements(); ) {
                root = set(root, e.nextInt());
            }
        }
        check(root);
        return root;
    }

    static String hex(int i) {
        return Integer.toHexString(i);
    }

    public int memoryUsage() {
        return memoryUsage(_root);
    }

    public static int memoryUsage(int root) {
        int n;
        switch (getType(root)) {
        case EMPTY_SET:
            return 0;
        case ONE_MEMBER_SET:
            return 0;
        case TWO_MEMBER_SET:
            return 4;
        case THREE_MEMBER_SET:
            return 8;
        case SMALL_SET:
            n = ((memory.get(root & MAX_BIT) & 0xffff) * 4) + 4;
            return n;
        case MEDIUM_SET:
            n = (memory.get((root & MAX_BIT) + 1) * 8) + 8;
            return n;
        case LARGE_SET:
            n = treeMemoryUsage(rootSubset, root & MAX_BIT) * 4;
            return n;
        default:
            assert(false);
        }
        return 0;
    }

    public int type() {
        return getType(_root);
    }

    public int numberOfElements() {
        return numberOfElements(_root);
    }

    public static int numberOfElements(int root) {
        switch (getType(root)) {
        case EMPTY_SET:
            return 0;
        case ONE_MEMBER_SET:
            return 1;
        case TWO_MEMBER_SET:
            return 2;
        case THREE_MEMBER_SET:
            return 3;
        case SMALL_SET:
            return  memory.get(root & MAX_BIT) & 0xffff;
        case MEDIUM_SET:
            int pop = 0;
            int r = root & MAX_BIT;
            int size = memory.get(r);
            int max = memory.get(r + 1);
            for (int i = 0; i < size; i++) {
                pop += memory.get(r + 2 + max + i);
            }
            return pop;
        case LARGE_SET:
            return treePopulation(rootSubset, root & MAX_BIT);
        default:
            assert(false);
        }
        return 0;
    }

    public int size() {
        assert(false);
        return 0;
    }

    public IntEnumeration elements() {
        return elements(_root);
    }

    public static IntEnumeration elements(int root) {
        return new BitEnum(root);
    }

    static class BitEnum implements IntEnumeration {
        int index;
        int offset;
        int root;

        BitEnum(int root) {
            this.root = root;
            reset();
        }

        public boolean hasMoreElements() {
            switch (getType(root)) {
            case EMPTY_SET:
                return false;
            case ONE_MEMBER_SET:
                return index == 0;
            case TWO_MEMBER_SET:
                return index <= 1;
            case THREE_MEMBER_SET:
                return index <= 2;
            case SMALL_SET:
                return index < (memory.get(root & MAX_BIT) & 0xffff);
            case MEDIUM_SET:
                return index < memory.get(root & MAX_BIT);
            case LARGE_SET:
                return index != -1;
            default:
                assert(false);
            }
            return false;
        }

        public Object nextElement() {
            return new Integer(nextInt());
        }

        public int nextInt() {
            int r = root & MAX_BIT;
            switch (getType(root)) {
            case ONE_MEMBER_SET:
                index++;
                return root & MAX_BIT;
            case TWO_MEMBER_SET:
            case THREE_MEMBER_SET:
                return memory.get(r + index++);
            case SMALL_SET:
                return memory.get(r + 1 + index++);
            case MEDIUM_SET:
                int i = memory.get(r + 2 + index) + offset;
                int max = memory.get(r + 1);
                if (++offset >= memory.get(r + 2 + max + index)) {
                    index++;
                    offset = 0;
                }
                return i;
            case LARGE_SET:
                int bit = index;
                index = treeNextBit(rootSubset, r, index + 1);
                return bit;
            default:
                assert(false);
            }
            return -1;
        }

        public int peekInt() {
            int r = root & MAX_BIT;
            switch (getType(root)) {
            case ONE_MEMBER_SET:
                return root & MAX_BIT;
            case TWO_MEMBER_SET:
            case THREE_MEMBER_SET:
                return memory.get(r + index);
            case SMALL_SET:
                return memory.get(r + 1 + index);
            case MEDIUM_SET:
                return memory.get(r + 2 + index) + offset;
            case LARGE_SET:
                return index;
            default:
                assert(false);
            }
            return -1;
        }

        public void reset() {
            index = 0;
            offset = 0;
            if (getType(root) == LARGE_SET) {
                index = treeNextBit(rootSubset, root & MAX_BIT, 0);
            }
        }
    }

    protected void finalize() throws Throwable {
        free(_root);
    }

    public static void free(int root) {
        int r = root & MAX_BIT;
        int max;
        switch (getType(root)) {
        case EMPTY_SET:
        case ONE_MEMBER_SET:
            return;
        case TWO_MEMBER_SET:
            freeroot(r, 1);
            return;
        case THREE_MEMBER_SET:
            freeroot(r, 2);
            return;
        case SMALL_SET:
            max = memory.get(r) >>> 16;
            freeroot(r, max + 1);
            return;
        case MEDIUM_SET:
            max = memory.get(r + 1);
            freeroot(r, (max * 2) + 2);
            return;
        case LARGE_SET:
            treeFree(rootSubset, r);
            liveRoots.clear(r);
            return;
        default:
            assert(false);
        }
    }

    static void treeWriteObject(ObjectOutputStream out, SubsetClass sc, int subsetRoot) 
                                        throws IOException {
        for (int i = 0; i < sc.subsets; i++) {
            int nextSubsetRoot = memory.get(subsetRoot + i);
            if (sc.next != null && nextSubsetRoot != EMPTY_TREE && nextSubsetRoot != FULL_TREE) {
                nextSubsetRoot &= ~COPY_ON_WRITE;
                out.writeInt(PARTIAL_TREE);
                treeWriteObject(out, sc.next, nextSubsetRoot);
            } else {
                out.writeInt(nextSubsetRoot);
            }
        }
    }

    static int treeReadObject(ObjectInputStream in, SubsetClass sc) throws IOException {
        int subsetRoot = treeAlloc(sc);
        for (int i = 0; i < sc.subsets; i++) {
            int nextSubsetRoot = in.readInt();
            if (sc.next != null && nextSubsetRoot == PARTIAL_TREE) {
                nextSubsetRoot = treeReadObject(in, sc.next);
            }
            memory.put(subsetRoot + i, nextSubsetRoot);
        }
        return subsetRoot;
    }

    private void writeObject(ObjectOutputStream out) throws IOException {
        writeRoot(out, _root);
    }

    public static void writeRoot(ObjectOutputStream out, int root) throws IOException {
        out.writeInt(root);
        int r = root & MAX_BIT;
        int max;
        switch (getType(root)) {
        case EMPTY_SET:
        case ONE_MEMBER_SET:
            return;
        case TWO_MEMBER_SET:
            out.writeInt(memory.get(r));
            out.writeInt(memory.get(r + 1));
            return;
        case THREE_MEMBER_SET:
            out.writeInt(memory.get(r));
            out.writeInt(memory.get(r + 1));
            out.writeInt(memory.get(r + 2));
            return;
        case SMALL_SET:
            max = memory.get(r) >>> 16;
            out.writeInt(memory.get(r));
            for (int i = 0; i < max; i++)
                out.writeInt(memory.get(r + 1 + i));
            return;
        case MEDIUM_SET:
            max = memory.get(r + 1);
            out.writeInt(memory.get(r));
            out.writeInt(memory.get(r + 1));
            for (int i = 0; i < max * 2; i++)
                out.writeInt(memory.get(r + 2 + i));
            return;
        case LARGE_SET:
            treeWriteObject(out, rootSubset, r);
            return;
        default:
            assert(false);
        }
    }

    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
        _root = readRoot(in);
    }

    public static int readRoot(ObjectInputStream in) throws IOException, ClassNotFoundException {
        int root = in.readInt();
        int type = getType(root);
        int tmp, size, max;
        switch (type) {
        case EMPTY_SET:
        case ONE_MEMBER_SET:
            return root;
        case TWO_MEMBER_SET:
            root = rootalloc(2);
            memory.put(root, in.readInt());
            memory.put(root + 1, in.readInt());
            return setType(root, type);
        case THREE_MEMBER_SET:
            root = rootalloc(3);
            memory.put(root, in.readInt());
            memory.put(root + 1, in.readInt());
            memory.put(root + 2, in.readInt());
            return setType(root, type);
        case SMALL_SET:
            tmp = in.readInt();
            size = tmp & 0xffff;
            max = tmp >>> 16;
            root = rootalloc(max + 1);
            memory.put(root, tmp);
            for (int i = 0; i < max; i++)
                memory.put(root + 1 + i, in.readInt());
            return setType(root, type);
        case MEDIUM_SET:
            size = in.readInt();
            max = in.readInt();
            root = rootalloc((max * 2) + 2);
            memory.put(root, size);
            memory.put(root + 1, max);
            for (int i = 0; i < max * 2; i++)
                memory.put(root + 2 + i, in.readInt());
            return setType(root, type);
        case LARGE_SET:
            root = treeReadObject(in, rootSubset);
            return setType(root, type);
        default:
            assert(false);
        }
        return 0;
    }

    static void assert(boolean condition) {
        if (!condition) throw new Error("assert failed!");
    }

    static class TestBits {
        BitSet bits;
        TreeBitSet set;

        TestBits(int min, int interval, int max) {
            set = new TreeBitSet();
            bits = new BitSet();
            for (int i = min; i < max; i += interval) {
                set.set(i);
                bits.set(i);
            }
        }

        void or(TestBits other) {
            set.or(other.set);
            bits.or(other.bits);
        }

        void expect(int min, int interval, int max) {
            int count = 0;
            for (int i = min; i < max; i += interval, count++) {
                if (!set.get(i))
                    throw new Error("i " + i + " was not set!");
            }
            if (set.numberOfElements() != count)
                throw new Error("expected " + count + " found " + set.numberOfElements());
            for (int i = max; i < max * 2; i++) {
                if (set.get(i))
                    throw new Error("i " + i + " IS set!");
            }
            BitSet nbits = new BitSet();
            for (IntEnumeration e = set.elements(); e.hasMoreElements(); ) {
                int i = e.nextInt();
                if (!bits.get(i))
                    throw new Error("i " + i + " was not set!");
                nbits.set(i);
            }
            if (!bits.equals(nbits))
                throw new Error("enum did not match");
        }
    }

    public static void main(String args[]) {
int f = treeAlloc(rootSubset);
System.out.println("usage = " + memory.size());
treeSet(rootSubset, f, 1);
System.out.println("usage = " + memory.size());
        TestBits a = new TestBits(0, 2, 2049);
        a.expect(0, 2, 2049);
        a = new TestBits(0, 1, 1000);
        a.expect(0, 1, 1000);
        a = new TestBits(0, 2, 1000);
        a.expect(0, 2, 1000);
        TestBits b = new TestBits(1, 2, 1000);
        b.expect(1, 2, 1000);
        a.or(b);
        a.expect(0, 1, 1000);
        b = new TestBits(1000, 2, 2000);
        b.expect(1000, 2, 2000);
        a.or(b);
        b.expect(1000, 2, 2000);
        b = new TestBits(1001, 2, 2000);
        a.or(b);
        b.expect(1001, 2, 2000);
        a.expect(0, 1, 2000);
        a = new TestBits(0, 2, 8000);
        b = new TestBits(1, 2, 2);
        b.or(a);
        a = new TestBits(3, 2, 8000);
        TestBits c = new TestBits(3, 2, 8000);
        c.or(b);
        c.expect(0, 1, 8000);
        c = new TestBits(3, 2, 4);
        c.or(b);
        c.or(new TestBits(0, 1, 3));
        TestBits d = new TestBits(3, 2, 4);
        d.or(b);
        c.or(d);
        c.or(new TestBits(3, 2, 8000));
        c.expect(0, 1, 8000);
        b.or(a);
        b.expect(0, 1, 8000);
        for (int i = 0; i < 1000; i++) {
            a = new TestBits(0, 100, i);
            a.expect(0, 100, i);
            a = new TestBits(0, 2, i);
            a.expect(0, 2, i);
            a = new TestBits(0, 1, i);
            a.expect(0, 1, i);
        }
    }
}

class SubsetClass {
    SubsetClass next;
    int subsets;
    int size;
    int shift;
    int mask;
    int bits;

    SubsetClass(SubsetClass next, int bits) {
        this.next = next;
        this.bits = bits;
        if (next == null) {
            size = 1 << bits;
            subsets = size / 32;
        } else if (next.next == null) {
            size = next.size;
            subsets = 1 << bits;
            shift = next.bits;
        } else {
            subsets = 1 << bits;
            shift = next.shift + next.bits;
            size = 1 << shift;
        }
        mask = size - 1;
    }

    int subsetIndex(int n) {
        return n >>> shift;
    }

    int subsetBits(int n) {
        return n & mask;
    }
}

final class Memory {
    static final int shift = 19;
    static final int CHUNK_SIZE = 1 << shift;
    static final int mask = CHUNK_SIZE - 1;
    int chunks[][] = new int[1][CHUNK_SIZE];
    int chunkOffset;

    Memory() {}

    void add(int value) {
        if (chunkOffset >= CHUNK_SIZE) {
            int newchunks[][] = new int[chunks.length + 1][];

            for (int i = 0; i < chunks.length; i++) {
                newchunks[i] = chunks[i];
            }
            newchunks[chunks.length] = new int[CHUNK_SIZE];
            chunks = newchunks;
            chunkOffset = 0;
        }
        chunks[chunks.length - 1][chunkOffset++] = value;
    }

    int size() {
        return ((chunks.length - 1) << shift) + chunkOffset;
    }

    final int get(int index) {
        return chunks[index >> shift][index & mask];
    }

    void put(int index, int value) {
        chunks[index >> shift][index & mask] = value;
    }
}
