package com.ibm.jvm.util.html;

import java.util.*;

/**
 *  This class represents a HTML table cell.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Cell extends Element {
    /** The contents of this cell. */
    Element contents;
    /** The column this cell belongs to. */
    Column column;
    /** Whether this is a header cell. */
    boolean isHeader;
    /** The width of this cell. */
    int width = -1;
    /** True if the width is a percentage */
    boolean usePercent;
    /** The background colour of this cell. */
    String bgcolor;

    /**
     * Creates a new Cell containing the given element.
     * @param contents the element to display in this cell
     */
    public Cell(Element contents) {
        this.contents = contents;
    }

    /**
     * Creates a new Cell containing the given element with width and background colour.
     * @param contents the element to display in this cell
     * @param width the width of the cell
     * @param bgcolor the background color of the cell
     */
    public Cell(Element contents, int width, String bgcolor) {
        this.contents = contents;
        this.width = width;
        this.bgcolor = bgcolor;
    }

    /**
     * Creates a new Cell containing the given element with width and background colour.
     * @param contents the element to display in this cell
     * @param width the width of the cell
     * @param usePercent true if the width is a percentage
     * @param bgcolor the background color of the cell
     */
    public Cell(Element contents, int width, boolean usePercent, String bgcolor) {
        this(contents, width, bgcolor);
        this.usePercent = true;
    }

    /**
     * Creates a new Cell containing the given text.
     * @param text the text to display in this cell
     */
    public Cell(String text) {
        contents = new Text(text);
    }

    /**
     * Creates a new Cell containing the given text and background colour.
     * @param text the text to display in this cell
     * @param bgcolor the background color of the cell
     */
    public Cell(String text, String bgcolor) {
        this(text);
        this.bgcolor = bgcolor;
    }

    /**
     * Sets whether this is a header cell.
     */
    public void setHeader(boolean flag) {
        isHeader = flag;
    }

    /**
     * Sets the column for this cell.
     * @param column the Column
     */
    public void setColumn(Column column) {
        this.column = column;
    }

    /**
     * Returns the column for this cell.
     */
    public Column column() {
        return column;
    }

    /**
     * Returns the text for this cell.
     */
    public String toString() {
        return contents.toString();
    }

    /**
     * Returns a html representation of this cell.
     */
    protected String toHtmlString() {
        char h = isHeader ? 'h' : 'd';
        String s = "<t" + h;
        if (width != -1) {
            s += " width=\"" + width;
            if (usePercent) {
                s += "%";
            }
            s += "\"";
        }
        if (bgcolor != null) {
            s += " bgcolor=\"" + bgcolor + "\"";
        }
        s += ">";
        s = beautify(s);
        String c = contents.toHtmlString();
        if (c.length() == 0) {
            c = "&nbsp;";
        }
        return beautify(s + c + "</t" + h + ">");
    }

    /**
     * Returns a text representation of this cell.
     */
    protected String toTextString() {
        return contents.toTextString();
    }

    /**
     * Returns the width in characters of this cell.
     */
    public int width() {
        return toTextString().length();
    }
}
