package com.ibm.jvm.util.html;

import java.util.*;

/**
 *  This class represents a HTML Document.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Document extends Base {
    /** Output a representation of the document in plain text. */
    protected boolean usePlainText;
    /** The list of top level elements in this document. */
    protected ArrayList elements = new ArrayList();
    /** The title string. */
    protected String title;
    /** The system line separator. */
    protected static final String lineSeparator = System.getProperty("line.separator");
    /** The margin width */
    protected int marginwidth = -1;
    /** The margin height */
    protected int marginheight = -1;

    /**
     * Create a new html document with the given title.
     * @param title the title of the document
     */
    public Document(String title) {
        this.title = title;
    }

    /**
     * Create a new html document with the given title and margin width and height.
     * @param title the title of the document
     * @param marginwidth the margin width
     * @param marginheight the margin height
     */
    public Document(String title, int marginwidth, int marginheight) {
        this.title = title;
        this.marginwidth = marginwidth;
        this.marginheight = marginheight;
    }

    /**
     * Create a new html document with no title.
     */
    public Document() {
        title = "";
    }

    /**
     * Add a new html element to this document. 
     */
    public void addElement(Element element) {
        element.setDocument(this);
        elements.add(element);
    }

    public void add(Element element) {
        addElement(element);
    }

    /**
     * Set whether or not we use plain text for output.
     */
    public void setPlainText(boolean plainText) {
        usePlainText = plainText;
    }

    /**
     * Adds a Text element to the document containing the given text.
     */
    public void print(String text) {
        addElement(new Text(text));
    }

    /**
     * Adds a Text element to the document containing the given text followed by a newline.
     */
    public void println(String text) {
        addElement(new Text(text, true));
    }

    /**
     * Adds a Text element to the document containing the given text as a header.
     * The level defaults to 2.
     * @param text the text to be added
     */
    public void printHeader(String text) {
        printHeader(text, 2);
    }

    /**
     * Adds a Text element to the document containing the given text as a header at the
     * givel level.
     * @param text the text to be added
     * @param level the header level
     */
    public void printHeader(String text, int level) {
        addElement(new Text(text, level));
    }

    /**
     * Returns a html representation of this document.
     */
    protected String toHtmlString() {
        String header = "<html><head><title>" + title + "</title></head>";
        String body = "<body";
        if (marginwidth != -1) {
            body += " marginwidth=\"" + marginwidth + "\"";
        }
        if (marginheight != -1) {
            body += " marginheight=\"" + marginheight + "\"";
        }
        body += ">";
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            Element element = (Element)i.next();
            body += element.toHtmlString();
        }
        return header + body + "</body></html>";
    }

    /**
     * Return a string containing a repeated character.
     * @param n the number of times to repeat
     * @param c the character
     */
    protected static String pad(int n, char c) {
        assert(n >= 0);
        String pad = "";
        for (int i = 0; i < n; i++) {
            pad += c;
        }
        return pad;
    }

    /**
     * Returns a text representation of this document.
     */
    protected String toTextString() {
        String str = "";
        for (Iterator i = elements.iterator(); i.hasNext(); ) {
            Element element = (Element)i.next();
            str += element.toTextString();
        }
        return str;
    }

    /** 
     * Returns a representation of this document.
     */
    public String toString() {
        if (usePlainText)
            return toTextString();
        else
            return toHtmlString();
    }
}
