package com.ibm.jvm.util.html;

import java.util.*;

/**
 *  This class represents a HTML table row.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Row extends Element {
    /** The containing Table. */
    protected Table table;
    /** The list of cells in this row. */
    protected ArrayList cells = new ArrayList();
    /** Indicates that this is a header row. */
    protected boolean isHeader;

    /**
     * Creates a new Row.
     */
    public Row() {
    }

    /**
     * Creates a new Row containing the given strings as cells.
     */
    public Row(String strings[]) {
        for (int i = 0; i < strings.length; i++) {
            addCell(strings[i]);
        }
    }

    /**
     * Creates a new Row with given header flag.
     * @param isHeader true if this is a header row
     */
    public Row(boolean isHeader) {
        this.isHeader = isHeader;
    }

    /**
     * Sets the header flag.
     * @param flag the new value of the header flag
     */
    public void setHeader(boolean flag) {
        isHeader = flag;
    }

    /**
     * Sets the containing Table.
     */
    public void setTable(Table table) {
        this.table = table;
    }

    /**
     * Adds a new cell to the row.
     * @param cell the cell to be added.
     */
    public void addCell(Cell cell) {
        if (table != null) {
            /*
             * If the row has already been added to the table we need to
             * add the cell to the table column here.
             */
            table.addCellToColumn(cell, cells.size());
        }
        cells.add(cell);
    }

    /**
     * Adds a new cell to the row.
     * @param cell String version of the cell to be added.
     */
    public void addCell(String cell) {
        assert(cell != null);
        addCell(new Cell(cell));
    }

    /**
     * Adds a new cell to the row.
     * @param cell Element version of the cell to be added.
     */
    public void addCell(Element cell) {
        addCell(new Cell(cell));
    }

    /**
     * Returns the number of cells in this row.
     */
    public int size() {
        return cells.size();
    }

    /**
     * Returns the cell at the given position.
     * @param index index of the cell to be returned
     */
    public Cell cellAt(int index) {
        return (Cell)cells.get(index);
    }

    /**
     * Returns a html representation of this row.
     */
    protected String toHtmlString() {
        String rep = beautify("<tr>");
        for (Iterator it = cells.iterator(); it.hasNext();) {
            Cell cell = (Cell)it.next();
            if (isHeader) {
                cell.setHeader(true);
            }
            rep += cell.toHtmlString();
        }
        rep += "</tr>";
        return beautify(rep);
    }

    /**
     * Returns a text representation of this row.
     */
    protected String toTextString() {
        String rep = "";
        for (Iterator it = cells.iterator(); it.hasNext();) {
            Cell cell = (Cell)it.next();
            rep += cell.toTextString();
            rep += Document.pad(cell.column().width() - cell.width(), ' ');
            rep += Document.pad(table.columnGap(), ' ');
        }
        rep += Document.lineSeparator;
        if (isHeader) {
            // Underline column headers
            for (Iterator it = cells.iterator(); it.hasNext();) {
                Cell cell = (Cell)it.next();
                rep += Document.pad(cell.width(), '-');
                rep += Document.pad(cell.column().width() - cell.width(), ' ');
                rep += Document.pad(table.columnGap(), ' ');
            }
            rep += Document.lineSeparator;
        }
        return rep;
    }
}
