package com.ibm.jvm.util.html;

import java.util.*;

/**
 *  This class represents a HTML Table.
 *
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Table extends Element {

    /** The header for this table. */
    protected Row headerRow;
    /** The list of rows for this table. */
    protected ArrayList rows = new ArrayList();
    /** The list of columns for this table. */
    protected ArrayList columns = new ArrayList();
    /** The gap in spaces between columns when in plain text mode. */
    protected int columnGap = 3;
    /** The cellspacing attribute to use. */
    protected int cellspacing = -1;
    /** The cellpadding attribute to use. */
    protected int cellpadding = 4;
    /** The border attribute to use. */
    protected int border = 1;
    /** The width attribute to use. */
    protected int width = -1;
    /** Whether the width is percentage */
    protected boolean usePercent;
    /** The background colour */
    protected String bgcolor;
    /** Whether to supress the output of empty columns. */
    protected boolean suppressEmptyColumns;

    /**
     * Creates a new Table.
     */
    public Table() {
    }

    /**
     * Creates a new Table with the given headers.
     * @param headers an array of Strings that specify the table headers
     */
    public Table(String headers[]) {
        createHeaderRow();
        for (int i = 0; i < headers.length; i++) {
            Cell cell = new Cell(headers[i]);
            headerRow.addCell(cell);
        }
    }

    public Table(String h1, String h2) {
        this(new String[] {h1, h2});
    }

    public Table(String h1, String h2, String h3) {
        this(new String[] {h1, h2, h3});
    }

    /**
     * Creates a new Table with the given border, spacing and padding.
     */
    public Table(int border, int cellspacing, int cellpadding) {
        this.border = border;
        this.cellspacing = cellspacing;
        this.cellpadding = cellpadding;
    }

    /**
     * Creates a new Table with the given border, spacing, padding and background colour.
     */
    public Table(int border, int cellspacing, int cellpadding, String bgcolor) {
        this(border, cellspacing, cellpadding);
        this.bgcolor = bgcolor;
    }

    /**
     * Creates a new Table with the given width, border, spacing, padding and background colour.
     * Note that the width defaults to percent unlike Cell!
     */
    public Table(int width, int border, int cellspacing, int cellpadding, String bgcolor) {
        this(border, cellspacing, cellpadding, bgcolor);
        this.width = width;
        usePercent = true;
    }

    /**
     * Sets the suppressEmptyColumns attribute.
     */
    public void suppressEmptyColumns(boolean suppressEmptyColumns) {
        this.suppressEmptyColumns = suppressEmptyColumns;
    }

    /**
     * Sets the cellspacing attribute.
     */
    public void setCellSpacing(int cellspacing) {
        this.cellspacing = cellspacing;
    }

    /**
     * Sets the cellpadding attribute.
     */
    public void setCellPadding(int cellpadding) {
        this.cellpadding = cellpadding;
    }

    /**
     * Sets the border attribute.
     */
    public void setBorder(int border) {
        this.border = border;
    }

    /**
     * Sets the column gap for plain text mode.
     * @param columnGap the number of spaces between columns
     */
    public void setColumnGap(int columnGap) {
        this.columnGap = columnGap;
    }

    /**
     * Returns the columnGap.
     */
    public int columnGap() {
        return columnGap;
    }

    /**
     * Creates the headerRow.
     */
    private void createHeaderRow() {
        headerRow = new Row(true);
        headerRow.setTable(this);
    }

    /**
     * Adds a cell to the header row.
     * @param string the text to be added
     */
    public void addHeaderCell(String string) {
        if (headerRow == null) {
            createHeaderRow();
        }
        Cell cell = new Cell(string);
        headerRow.addCell(cell);
    }

    /**
     * Adds a new cell to the given column.
     * @param cell the cell to add
     * @param colindex the index of the column to add it to
     */
    public void addCellToColumn(Cell cell, int colindex) {
        Column column;
        if (colindex == columns.size()) {
            column = new Column();
            addColumn(column);
        } else {
            column = (Column)columns.get(colindex);
        }
        column.addCell(cell);
    }

    /**
     * Adds a new column.
     * @param column the column to add
     */
    public void addColumn(Column column) {
        columns.add(column);
    }

    /**
     * Adds a new row.
     * @param row the row to add
     */
    public void addRow(Row row) {
        rows.add(row);
        row.setTable(this);
        for (int i = 0; i < row.size(); i++) {
            Cell cell = row.cellAt(i);
            addCellToColumn(cell, i);
        }
    }

    /**
     * Adds a new row formed from the given array of strings.
     */
    public void addRow(String strings[]) {
        Row row = new Row(strings);
        addRow(row);
    }

    public void addRow(String c1, String c2) {
        addRow(new String[] {c1, c2});
    }

    public void addRow(String c1, String c2, String c3) {
        addRow(new String[] {c1, c2, c3});
    }

    /**
     * Returns a html representation of this table.
     */
    protected String toHtmlString() {
        String rep = "<table border=\"" + border + "\" cellpadding=\"" + cellpadding + "\"";
        if (cellspacing != -1) {
            rep += " cellspacing=\"" + cellspacing + "\"";
        }
        if (bgcolor != null) {
            rep += " bgcolor=\"" + bgcolor + "\"";
        }
        if (width != -1) {
            rep += " width=\"" + width;
            if (usePercent) {
                rep += "%";
            }
            rep += "\"";
        }
        rep += ">";
        rep = beautify(rep);
        if (headerRow != null) {
            rep += headerRow.toHtmlString();
        }
        for (Iterator it = rows.iterator(); it.hasNext();) {
            Row row = (Row)it.next();
            rep += row.toHtmlString();
        }
        rep += "</table>";
        return beautify(rep);
    }

    /**
     * Returns a text representation of this table.
     */
    protected String toTextString() {
        String rep = "";
        if (headerRow != null) {
            rep += headerRow.toTextString();
        }
        for (Iterator it = rows.iterator(); it.hasNext();) {
            Row row = (Row)it.next();
            rep += row.toTextString();
        }
        return rep;
    }
}
