package com.ibm.jvm.zseries;


import com.ibm.jvm.util.*;
import com.ibm.jvm.util.html.*;
import java.util.*;
import java.io.*;


/**
 *  @author David Griffiths <tt>&lt;dave@prim.demon.co.uk&gt;</tt>
 */

public class Engine {

    RandomAccessStream ras;
    boolean r15PointsToEntry = true;
    int reg[] = new int[16];
    boolean regValid[] = new boolean[16];

    public Engine(RandomAccessStream ras) {
        this.ras = ras;
    }

    int getRegister(int r) {
        return reg[r];
    }

    public void putRegister(int r, int value) {
        //System.out.println("putRegister $" + r + " = " + hx(value) + " (was " + hx(reg[r]) + ")");
        reg[r] = value;
        regValid[r] = true;
    }

    public boolean validRegister(int r) {
        return regValid[r];
    }

    public void killRegister(int r) {
        regValid[r] = false;
    }

    public Instruction[] disassemble(Document doc, int address, int count, int caa) throws Exception {
        putRegister(12, caa);
        return disassemble(doc, address, count);
    }

    public Instruction[] disassemble(Document doc, int address, int count) throws Exception {
        int startAddress = address;

        putRegister(15, address);
        Vector v = new Vector();
        boolean foundStart = false;
        Instruction inst = null;

        for (int i = 0; i < count; i++) {
            int word = ras.readInt(address);
            int b = word >>> 24;
            boolean recognized = true;

            switch (b) {
            case 0x05:
                inst = new BranchAndLink(this, word);
                break;

            case 0x06:
                inst = new BranchOnCountR(this, word);
                break;

            case 0x07:
                inst = new BranchOnConditionR(this, word);
                break;

            case 0x0a:
                inst = new SupervisorCall(this, word);
                break;

            case 0x0d:
                inst = new BranchAndSaveR(this, word);
                break;

            case 0x12:
                inst = new LoadAndTest(this, word);
                break;

            case 0x13:
                inst = new LoadComplement(this, word);
                break;

            case 0x15:
                inst = new CompareLogicalR(this, word);
                break;

            case 0x17:
                inst = new XorR(this, word);
                break;

            case 0x18:
                inst = new LoadRegister(this, word);
                break;

            case 0x19:
                inst = new CompareR(this, word);
                break;

            case 0x1a:
                inst = new AddR(this, word);
                break;

            case 0x1b:
                inst = new SubtractR(this, word);
                break;

            case 0x1e:
                inst = new AddLogicalR(this, word);
                break;

            case 0x1f:
                inst = new SubtractLogicalR(this, word);
                break;

            case 0x40:
                inst = new StoreHalfword(this, word);
                break;

            case 0x41:
                inst = new LoadAddress(this, word);
                break;

            case 0x42:
                inst = new StoreCharacter(this, word);
                break;

            case 0x43:
                inst = new InsertCharacter(this, word);
                break;

            case 0x44:
                inst = new Execute(this, word);
                break;

            case 0x46:
                inst = new BranchOnCount(this, word);
                break;

            case 0x47:
                inst = new BranchOnCondition(this, word);
                if (i == 0 && word == 0x47f0f001) {
                    recognized = false;
                }
                break;

            case 0x48:
                inst = new LoadHalfword(this, word);
                break;

            case 0x49:
                inst = new CompareHalfword(this, word);
                break;

            case 0x4a:
                inst = new AddHalfword(this, word);
                break;

            case 0x4d:
                inst = new BranchAndSave(this, word);
                break;

            case 0x50:
                inst = new Store(this, word);
                break;

            case 0x54:
                inst = new And(this, word);
                break;

            case 0x55:
                inst = new CompareLogical(this, word);
                break;

            case 0x56:
                inst = new Or(this, word);
                break;

            case 0x57:
                inst = new Xor(this, word);
                break;

            case 0x58:
                inst = new Load(this, word);
                break;

            case 0x59:
                inst = new Compare(this, word);
                break;

            case 0x5a:
                inst = new Add(this, word);
                break;

            case 0x5b:
                inst = new Subtract(this, word);
                break;

            case 0x5d:
                inst = new Divide(this, word);
                break;

            case 0x5e:
                inst = new AddLogical(this, word);
                break;

            case 0x5f:
                inst = new SubtractLogical(this, word);
                break;

            case 0x60:
                inst = new StoreLong(this, word);
                break;

            case 0x68:
                inst = new LoadLong(this, word);
                break;

            case 0x70:
                inst = new StoreShort(this, word);
                break;

            case 0x78:
                inst = new LoadShort(this, word);
                break;

            case 0x88:
                inst = new ShiftRightSingleLogical(this, word);
                break;

            case 0x89:
                inst = new ShiftLeftSingleLogical(this, word);
                break;

            case 0x8a:
                inst = new ShiftRightSingle(this, word);
                break;

            case 0x8e:
                inst = new ShiftRightDouble(this, word);
                break;

            case 0x90:
                inst = new StoreMultiple(this, word);
                break;

            case 0x91:
                inst = new TestUnderMask(this, word);
                break;

            case 0x92:
                inst = new Move(this, word);
                break;

            case 0x94:
                inst = new AndImmediate(this, word);
                break;

            case 0x95:
                inst = new CompareLogicalImmediate(this, word);
                break;

            case 0x96:
                inst = new OrImmediate(this, word);
                break;

            case 0x98:
                inst = new LoadMultiple(this, word);
                break;

            case 0xa7:
                int branch = (word >> 16) & 0xf;

                if (branch == 0x0)
                    inst = new TestUnderMaskHigh(this, word);
                else if (branch == 0x1)
                    inst = new TestUnderMaskLow(this, word);
                else if (branch == 0x4)
                    inst = new BranchRelativeOnCondition(this, address, word);
                else if (branch == 0x8)
                    inst = new LoadHalfwordImmediate(this, word);
                else if (branch == 0xe)
                    inst = new CompareHalfwordImmediate(this, word);
                else if (branch == 0xa)
                    inst = new AddHalfwordImmediate(this, word);
                else
                    throw new Exception("Unknown instruction 0x" + hex(word) + " at 0x" + hex(address));
                break;

            case 0xb3:
                if ((word & 0xffff0000) == 0xb3840000)
                    inst = new SetFPC(this, word);
                else
                    throw new Exception("Unknown instruction 0x" + hex(word) + " at 0x" + hex(address));
                break;

            case 0xba:
                inst = new CompareAndSwap(this, word);
                break;

            case 0xbd:
                inst = new CompareCharactersUnderMask(this, word);
                break;

            case 0xbf:
                inst = new InsertCharactersUnderMask(this, word);
                break;

            case 0xc0:
                b = (word >> 16) & 0xf;
                if (b == 0) {
                    int r1 = (word >> 20) & 0xf;
                    int i2 = ras.readInt(address + 2);

                    inst = new LoadAddressRelativeLong(this, address, r1, i2);
                } else
                    throw new Exception("Unknown instruction 0x" + hex(word) + " at 0x" + hex(address));
                break;

            case 0xd2:
                inst = new MoveC(this, word, ras.readInt(address + 4));
                break;

            case 0xd5:
                inst = new CompareLogicalCharacter(this, word, ras.readInt(address + 4));
                break;

            case 0xd7:
                inst = new XorC(this, word, ras.readInt(address + 4));
                break;

            case 0xdd:
                inst = new TranslateAndTest(this, word, ras.readInt(address + 4));
                break;

            default:
                if (!foundStart && (b == 0 || b == 1)) {
                    inst = new Unknown(this, word);
                    recognized = false;
                } else if (b == 0 && inst instanceof BranchAndLink) {
                    inst = new Unknown(this, word);
                } else {
                    if (b == 0)
                        inst = new Unknown(this, word);
                    else
                        throw new Exception("Unknown instruction 0x" + hex(word) + " at 0x" + hex(address));
                }
                break;
            }
            if (recognized)
                foundStart = true;
            System.out.println("0x" + hex(address) + ": " + "(0x" + hex(address - startAddress) + "): " + inst);
            address = inst.nextAddress(address);
        }
        return (Instruction[]) v.toArray(new Instruction[0]);
    }

    static String hex(int word) {
        String s = Integer.toHexString(word);

        for (int i = s.length(); i < 8; i++)
            s = "0" + s;
        return s;
    }

    static String hx(int word) {
        return Integer.toHexString(word);
    }
}
